/* 
 * This is a modified version of:
 *
 * tkCanvas.c --
 *
 * Modifications are copyright (c) 1994 Malcolm Beattie.
 * You may distribute under the terms of the GNU General Public License,
 * as specified in the README file.
 *
 *	This module implements canvas widgets for the Tk toolkit.
 *	A canvas displays a background and a collection of graphical
 *	objects such as rectangles, lines, and texts.
 *
 * Copyright (c) 1991-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char rcsid[] = "$Id: tkpCanvas.c,v 1.2 1994/09/21 14:39:01 mbeattie Exp $";
#endif

/* $Log: tkpCanvas.c,v $
 * Revision 1.2  1994/09/21  14:39:01  mbeattie
 * Fixed IX_COORDS loop iteration.
 *
 * Revision 1.1  1994/09/05  16:03:30  mbeattie
 * Initial revision
 *
 */

#include "default.h"
#include "tkInt.h"
#include "tkConfig.h"
#include "tkperl.h"
#include "tkpCanvas.h"

WidgetClass canvasClassRec;
I32 canvasMethods[] =
{
    IX_ADDTAG, IX_BBOX, IX_BIND, IX_CANVASX, IX_CANVASY, IX_CONFIGURE,
    IX_COORDS, IX_CREATE, IX_DCHARS, IX_DELETE, IX_DTAG, IX_FIND, IX_FOCUS,
    IX_GETTAGS, IX_ICURSOR, IX_INDEX, IX_INSERT, IX_ITEMCONFIGURE, IX_LOWER,
    IX_MOVE, IX_POSTSCRIPT, IX_RAISE, IX_SCALE, IX_SCANMARK, IX_SCANDRAGTO,
    IX_SELECTADJUST, IX_SELECTCLEAR, IX_SELECTFROM, IX_SELECTITEM,
    IX_SELECTTO, IX_TYPE, IX_XVIEW, IX_YVIEW, IX_EOL
};


/*
 * See tkCanvas.h for key data structures used to implement canvases.
 */

/*
 * The structure defined below is used to keep track of a tag search
 * in progress.  Only the "prevPtr" field should be accessed by anyone
 * other than StartTagSearch and NextItem.
 */

typedef struct TagSearch {
    Tk_Canvas *canvasPtr;	/* Canvas widget being searched. */
    Tk_Uid tag;			/* Tag to search for.   0 means return
				 * all items. */
    Tk_Item *prevPtr;		/* Item just before last one found (or NULL
				 * if last one found was first in the item
				 * list of canvasPtr). */
    Tk_Item *currentPtr;	/* Pointer to last item returned. */
    int searchOver;		/* Non-zero means NextItem should always
				 * return NULL. */
} TagSearch;

/*
 * Information used for argv parsing.
 */


static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	DEF_CANVAS_BG_COLOR, Tk_Offset(Tk_Canvas, bgBorder),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	DEF_CANVAS_BG_MONO, Tk_Offset(Tk_Canvas, bgBorder),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	DEF_CANVAS_BORDER_WIDTH, Tk_Offset(Tk_Canvas, borderWidth), 0},
    {TK_CONFIG_DOUBLE, "-closeenough", "closeEnough", "CloseEnough",
	DEF_CANVAS_CLOSE_ENOUGH, Tk_Offset(Tk_Canvas, closeEnough), 0},
    {TK_CONFIG_BOOLEAN, "-confine", "confine", "Confine",
	DEF_CANVAS_CONFINE, Tk_Offset(Tk_Canvas, confine), 0},
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	DEF_CANVAS_CURSOR, Tk_Offset(Tk_Canvas, cursor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	DEF_CANVAS_HEIGHT, Tk_Offset(Tk_Canvas, height), 0},
    {TK_CONFIG_BORDER, "-insertbackground", "insertBackground", "Foreground",
	DEF_CANVAS_INSERT_BG, Tk_Offset(Tk_Canvas, insertBorder), 0},
    {TK_CONFIG_PIXELS, "-insertborderwidth", "insertBorderWidth", "BorderWidth",
	DEF_CANVAS_INSERT_BD_COLOR, Tk_Offset(Tk_Canvas, insertBorderWidth),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_PIXELS, "-insertborderwidth", "insertBorderWidth", "BorderWidth",
	DEF_CANVAS_INSERT_BD_MONO, Tk_Offset(Tk_Canvas, insertBorderWidth),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_INT, "-insertofftime", "insertOffTime", "OffTime",
	DEF_CANVAS_INSERT_OFF_TIME, Tk_Offset(Tk_Canvas, insertOffTime), 0},
    {TK_CONFIG_INT, "-insertontime", "insertOffTime", "OnTime",
	DEF_CANVAS_INSERT_ON_TIME, Tk_Offset(Tk_Canvas, insertOffTime), 0},
    {TK_CONFIG_PIXELS, "-insertwidth", "insertWidth", "InsertWidth",
	DEF_CANVAS_INSERT_WIDTH, Tk_Offset(Tk_Canvas, insertWidth), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	DEF_CANVAS_RELIEF, Tk_Offset(Tk_Canvas, relief), 0},
    {TK_CONFIG_PIXELS, "-scrollincrement", "scrollIncrement", "ScrollIncrement",
	DEF_CANVAS_SCROLL_INCREMENT, Tk_Offset(Tk_Canvas, scrollIncrement), 0},
    {TK_CONFIG_STRING, "-scrollregion", "scrollRegion", "ScrollRegion",
	DEF_CANVAS_SCROLL_REGION, Tk_Offset(Tk_Canvas, regionString),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground", "Foreground",
	DEF_CANVAS_SELECT_COLOR, Tk_Offset(Tk_Canvas, selBorder),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground", "Foreground",
	DEF_CANVAS_SELECT_MONO, Tk_Offset(Tk_Canvas, selBorder),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_PIXELS, "-selectborderwidth", "selectBorderWidth", "BorderWidth",
	DEF_CANVAS_SELECT_BD_COLOR, Tk_Offset(Tk_Canvas, selBorderWidth),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_PIXELS, "-selectborderwidth", "selectBorderWidth", "BorderWidth",
	DEF_CANVAS_SELECT_BD_MONO, Tk_Offset(Tk_Canvas, selBorderWidth),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground", "Background",
	DEF_CANVAS_SELECT_FG_COLOR, Tk_Offset(Tk_Canvas, selFgColorPtr),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground", "Background",
	DEF_CANVAS_SELECT_FG_MONO, Tk_Offset(Tk_Canvas, selFgColorPtr),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	DEF_CANVAS_WIDTH, Tk_Offset(Tk_Canvas, width), 0},
    {TKP_CONFIG_METHOD, "-xscrollmethod", "xScrollMethod", "ScrollMethod",
	0, Tk_Offset(Tk_Canvas, xScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_SLAVE, "-xscrollslave", "xScrollSlave", "ScrollSlave",
	0, Tk_Offset(Tk_Canvas, xScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_METHOD, "-yscrollmethod", "yScrollMethod", "ScrollMethod",
	0, Tk_Offset(Tk_Canvas, yScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_SLAVE, "-yscrollslave", "yScrollSlave", "ScrollSlave",
	0, Tk_Offset(Tk_Canvas, yScrollCallback), TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * List of all the item types known at present:
 */

static Tk_ItemType *typeList = NULL;	/* NULL means initialization hasn't
					 * been done yet. */

/*
 * Standard item types provided by Tk:
 */

extern Tk_ItemType TkArcType, TkBitmapType, TkLineType;
extern Tk_ItemType TkOvalType, TkPolygonType;
extern Tk_ItemType TkRectangleType, TkTextType, TkWindowType;

/*
 * Various Tk_Uid's used by this module (set up during initialization):
 */

static Tk_Uid allUid = NULL;
static Tk_Uid currentUid = NULL;

/*
 * Statistics counters:
 */

static int numIdSearches;
static int numSlowSearches;

/*
 * Prototypes for procedures defined later in this file:
 */

STATIC void		CanvasBindProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
STATIC void		CanvasBlinkProc _ANSI_ARGS_((ClientData clientData));
STATIC void		CanvasDoEvent _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    XEvent *eventPtr));
STATIC void		CanvasEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
STATIC int		CanvasFetchSelection _ANSI_ARGS_((
			    ClientData clientData, int offset,
			    char *buffer, int maxBytes));
STATIC void		CanvasFocusProc _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    int gotFocus));
STATIC void		CanvasLostSelection _ANSI_ARGS_((
			    ClientData clientData));
STATIC void		CanvasSelectTo _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    Tk_Item *itemPtr, int index));
STATIC void		CanvasSetOrigin _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    int xOrigin, int yOrigin));
STATIC int		CanvasTagsParseProc _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, Tk_Window tkwin, char *value,
			    char *widgRec, int offset));
STATIC char *		CanvasTagsPrintProc _ANSI_ARGS_((ClientData clientData,
			    Tk_Window tkwin, char *widgRec, int offset,
			    Tcl_FreeProc **freeProcPtr));
STATIC void		CanvasUpdateScrollbars _ANSI_ARGS_((
			    Tk_Canvas *canvasPtr));
STATIC void		TkpConfigureCanvas _ANSI_ARGS_((Tk_Canvas *canvasPtr,
                            int sp, int items, int flags));
STATIC void		DestroyCanvas _ANSI_ARGS_((ClientData clientData));
STATIC void		DisplayCanvas _ANSI_ARGS_((ClientData clientData));
STATIC void		TkpAddTag _ANSI_ARGS_((Tk_Item *itemPtr, Tk_Uid tag));
STATIC void		EventuallyRedrawArea _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    int x1, int y1, int x2, int y2));
STATIC int		TkpFindItems _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    int sp, int items, char *newTag));
STATIC int		TkpFindArea _ANSI_ARGS_((Tk_Canvas *canvasPtr,
                            int sp, double *rect, Tk_Uid uid, int enclosed));
STATIC double		GridAlign _ANSI_ARGS_((double coord, double spacing));
STATIC void		InitCanvas _ANSI_ARGS_((void));
STATIC Tk_Item *	NextItem _ANSI_ARGS_((TagSearch *searchPtr));
STATIC void		PickCurrentItem _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    XEvent *eventPtr));
STATIC void		RelinkItems _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    char *tag, Tk_Item *prevPtr));
STATIC Tk_Item *	StartTagSearch _ANSI_ARGS_((Tk_Canvas *canvasPtr,
			    char *tag, TagSearch *searchPtr));

/*
 * Custom option for handling "-tags" options for canvas items:
 */

Tk_CustomOption tkCanvasTagsOption = {
    CanvasTagsParseProc,
    CanvasTagsPrintProc,
    (ClientData) NULL
};

/*
 *--------------------------------------------------------------
 *
 * TkpCanvasCmd --
 *
 *	This procedure is invoked to process the "canvas"
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

I32 TkpCanvasCmd(ix, sp, items)
int ix;
int sp;
int items;     
{
    SV **st = stack_base + sp;
    register Tk_Canvas *canvasPtr;
    char *pathname;
    
    Tk_Window new;

    if (typeList == NULL) {
	InitCanvas();
    }

    if (items < 1)
        croak("wrong # args for canvas");

    pathname = TkpGenPathname(st[1]);
    new = Tk_CreateWindowFromPath(tclinterp, mainwin, pathname, (char *) NULL);
    if (!new)
	croak("failed to create window");

    /*
     * Initialize fields that won't be initialized by TkpConfigureCanvas,
     * or which TkpConfigureCanvas expects to have reasonable values
     * (e.g. resource pointers).
     */

    canvasPtr = (Tk_Canvas *) ckalloc(sizeof(Tk_Canvas));
    canvasPtr->tkwin = new;
    canvasPtr->display = Tk_Display(new);
    canvasPtr->interp = tclinterp;
    canvasPtr->firstItemPtr = NULL;
    canvasPtr->lastItemPtr = NULL;
    canvasPtr->borderWidth = 0;
    canvasPtr->bgBorder = NULL;
    canvasPtr->relief = TK_RELIEF_FLAT;
    canvasPtr->pixmapGC = None;
    canvasPtr->width = None;
    canvasPtr->height = None;
    canvasPtr->confine = 0;
    canvasPtr->selBorder = NULL;
    canvasPtr->selBorderWidth = 0;
    canvasPtr->selFgColorPtr = NULL;
    canvasPtr->selItemPtr = NULL;
    canvasPtr->selectFirst = -1;
    canvasPtr->selectLast = -1;
    canvasPtr->anchorItemPtr = NULL;
    canvasPtr->selectAnchor = 0;
    canvasPtr->insertBorder = NULL;
    canvasPtr->insertWidth = 0;
    canvasPtr->insertBorderWidth = 0;
    canvasPtr->insertOnTime = 0;
    canvasPtr->insertOffTime = 0;
    canvasPtr->insertBlinkHandler = (Tk_TimerToken) NULL;
    canvasPtr->focusItemPtr = NULL;
    canvasPtr->xOrigin = canvasPtr->yOrigin = 0;
    canvasPtr->drawableXOrigin = canvasPtr->drawableYOrigin = 0;
    canvasPtr->bindingTable = NULL;
    canvasPtr->currentItemPtr = NULL;
    canvasPtr->closeEnough = 0.0;
    canvasPtr->pickEvent.type = LeaveNotify;
    canvasPtr->state = 0;
    Zero(&canvasPtr->xScrollCallback, 1, struct tkpcallback);
    Zero(&canvasPtr->yScrollCallback, 1, struct tkpcallback);
    canvasPtr->scrollX1 = 0;
    canvasPtr->scrollY1 = 0;
    canvasPtr->scrollX2 = 0;
    canvasPtr->scrollY2 = 0;
    canvasPtr->regionString = NULL;
    canvasPtr->scrollIncrement = 1;
    canvasPtr->scanX = 0;
    canvasPtr->scanXOrigin = 0;
    canvasPtr->scanY = 0;
    canvasPtr->scanYOrigin = 0;
    canvasPtr->hotPtr = NULL;
    canvasPtr->hotPrevPtr = NULL;
    canvasPtr->cursor = None;
    canvasPtr->pixelsPerMM = WidthOfScreen(Tk_Screen(new));
    canvasPtr->pixelsPerMM /= WidthMMOfScreen(Tk_Screen(new));
    canvasPtr->flags = 0;
    canvasPtr->nextId = 1;

    Tk_SetClass(canvasPtr->tkwin, "Canvas");
    Tk_CreateEventHandler(canvasPtr->tkwin,
	    ExposureMask|StructureNotifyMask|FocusChangeMask,
	    CanvasEventProc, (ClientData) canvasPtr);
    Tk_CreateEventHandler(canvasPtr->tkwin, KeyPressMask|KeyReleaseMask
	    |ButtonPressMask|ButtonReleaseMask|EnterWindowMask
	    |LeaveWindowMask|PointerMotionMask, CanvasBindProc,
	    (ClientData) canvasPtr);
    Tk_CreateSelHandler(canvasPtr->tkwin, XA_STRING, CanvasFetchSelection,
	    (ClientData) canvasPtr, XA_STRING);
    TkpCreateWidgetHash(canvasPtr->tkwin, (ClientData) canvasPtr);
    TkpConfigureCanvas(canvasPtr, sp + 1, items - 1, 0);
    /*
     * If the above fails, we ought to Tk_DestroyWindow(canvasPtr->tkwin)
     * but I've mucked up error propagation. Oops.
     */

    st[0] = sv_2mortal(NewWidgetRef(Tk_PathName(canvasPtr->tkwin),
                                    canvasClassRec.stash));
    return sp;
}

/*
 *--------------------------------------------------------------
 *
 * TkpCanvasWidgetCmd --
 *
 *	This procedure is invoked to process the command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

I32 TkpCanvasWidgetCmd(ix, sp, items)
int ix;
int sp;
int items;
{
    SV **st = stack_base + sp;
    Tk_Canvas *canvasPtr;
    char c;
    Tk_Item *itemPtr = NULL;		/* Initialization needed only to
					 * prevent compiler warning. */
    TagSearch search;

    if (items < 1)
        croak("wrong # args");
    if (!WrefToWidget(st[1], 0, (ClientData *) &canvasPtr, "Canvas", 6))
        croak("non-widget argument passed");

    Tk_Preserve((ClientData) canvasPtr);

    st[0] = sv_mortalcopy(&sv_undef); /* default return value */
    
    switch (ix)
    {
      case IX_ADDTAG:
	if (items < 3)
	    croak("wrong # args for addtag");

	sp = TkpFindItems(canvasPtr, sp + 1, items - 1, SvPV(st[2], na));
	break;
	
      case IX_BBOX:
	{
	    int i, gotAny;
	    int x1 = 0, y1 = 0, x2 = 0, y2 = 0;	/* Initializations needed
						 * only to prevent compiler
						 * warnings. */

	    if (items < 2)
		croak("wrong # args for bbox");

	    gotAny = 0;
	    for (i = 2; i <= items; i++) {
		itemPtr = StartTagSearch(canvasPtr, SvPV(st[i], na), &search);
		for (; itemPtr != NULL; itemPtr = NextItem(&search)) {
		    if (!gotAny) {
			x1 = itemPtr->x1;
			y1 = itemPtr->y1;
			x2 = itemPtr->x2;
			y2 = itemPtr->y2;
			gotAny = 1;
		    } else {
			if (itemPtr->x1 < x1) {
			    x1 = itemPtr->x1;
			}
			if (itemPtr->y1 < y1) {
			    y1 = itemPtr->y1;
			}
			if (itemPtr->x2 > x2) {
			    x2 = itemPtr->x2;
			}
			if (itemPtr->y2 > y2) {
			    y2 = itemPtr->y2;
			}
		    }
		}
	    }
	    if (gotAny)
	    {
		av_store(stack, sp++, sv_2mortal(newSViv(x1)));
		av_store(stack, sp++, sv_2mortal(newSViv(y1)));
		av_store(stack, sp++, sv_2mortal(newSViv(x2)));
		av_store(stack, sp, sv_2mortal(newSViv(y2)));
	    }
	    break;
	}

      case IX_BIND:
	{
	    ClientData object;
	    char *tagorid;
	    
	    if (items < 2 || items > 5)
		croak("wrong # args for tkbind");

	    /*
	     * Figure out what object to use for the binding (individual
	     * item vs. tag).
	     */
	    
	    object = 0;
	    tagorid = SvPV(st[2], na);
	    
	    if (isdigit(UCHAR(tagorid[0]))) {
		int id;
		char *end;
		
		id = strtoul(tagorid, &end, 0);
		if (*end != 0) {
		    goto bindByTag;
		}
		for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
		     itemPtr = itemPtr->nextPtr) {
		    if (itemPtr->id == id) {
			object = (ClientData) itemPtr;
			break;
		    }
		}
		if (object == 0)
		    croak("item does not exist");
	    } else {
	      bindByTag:
		object = (ClientData) Tk_GetUid(tagorid);
	    }
	    
	    /*
	     * Make a binding table if the canvas doesn't already have
	     * one.
	     */
	    
	    if (canvasPtr->bindingTable == NULL) {
		canvasPtr->bindingTable = TkpCreateBindingTable(tclinterp);
	    }
	    
	    if (items > 3)
	    {
		unsigned long mask;
		
		if (st[4] == &sv_undef)
		{
		    if (Tk_DeleteBinding(tclinterp, canvasPtr->bindingTable,
					 object, SvPV(st[3], na)) != TCL_OK)
			croak("bad binding to delete");
		    goto done;
		}
		
		mask = TkpCreateBinding(canvasPtr->bindingTable, object,
					SvPV(st[3], na), st[4],
					(items == 5) ? st[5] : Nullsv);
		if (mask == 0)
		    croak("binding failed in canvas tkbind");
		if (mask & ~(ButtonMotionMask|Button1MotionMask
			     |Button2MotionMask|Button3MotionMask
			     |Button4MotionMask|Button5MotionMask
			     |ButtonPressMask|ButtonReleaseMask
			     |EnterWindowMask|LeaveWindowMask|KeyPressMask
			     |KeyReleaseMask|PointerMotionMask)) {
		    Tk_DeleteBinding(tclinterp, canvasPtr->bindingTable,
				     object, SvPV(st[3], na));
		    croak("requested illegal events");
		}
	    } else if (items == 3) {
		croak("returning binding info not yet implemented");
	    } else {
		croak("returning all binding info not yet implemented");
	    }
	    break;
	}
	
      case IX_CANVASX:
	{
	    int x;
	    double grid;

	    if (items < 2 || items > 3)
		croak("wrong # args for canvasx");

	    if (Tk_GetPixels(tclinterp, canvasPtr->tkwin,
			     SvPV(st[2], na), &x) != TCL_OK)
		croak("bad screenx in canvasx");

	    if (items == 3)
	    {
		if (TkGetCanvasCoord(canvasPtr, SvPV(st[3], na),
				     &grid) != TCL_OK)
		    croak("bad grid distance in canvasx");
	    }
	    else
		grid = 0.0;

	    x += canvasPtr->xOrigin;
	    st[0] = sv_2mortal(newSVnv(GridAlign((double) x, grid)));
	    break;
	}
	
      case IX_CANVASY:
	{
	    int y;
	    double grid;

	    if (items < 2 || items > 3)
		croak("wrong # args for canvasy");

	    if (Tk_GetPixels(tclinterp, canvasPtr->tkwin,
			     SvPV(st[2], na), &y) != TCL_OK)
		croak("bad screenx in canvasy");

	    if (items == 3)
	    {
		if (TkGetCanvasCoord(canvasPtr, SvPV(st[3], na),
				     &grid) != TCL_OK)
		    croak("bad grid distance in canvasy");
	    }
	    else
		grid = 0.0;

	    y += canvasPtr->yOrigin;
	    st[0] = sv_2mortal(newSVnv(GridAlign((double) y, grid)));
	    break;
	}

      case IX_CONFIGURE:
	if (items == 1)
	    sp = TkpConfigureInfo(sp, canvasPtr->tkwin, configSpecs,
				  (char *) canvasPtr, (char *) NULL, 0);
	else if (items == 2)
	    sp = TkpConfigureInfo(sp, canvasPtr->tkwin, configSpecs,
				  (char *) canvasPtr, SvPV(st[2], na), 0);
	else
	    TkpConfigureCanvas(canvasPtr, sp + 1, items - 1,
			       TK_CONFIG_ARGV_ONLY);

	break;

      case IX_COORDS:
	if (items < 2)
	    croak("wrong # args for coords");
	
	itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
	if (itemPtr != NULL) {
	    if (items != 2)
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
	    if (itemPtr->typePtr->coordProc != NULL)
	    {
		Tcl_ResetResult(tclinterp);
		if (items != 2)
		{
		    char **argv;
		    int result;
		    
		    argv = StackToArgv(sp + 3, items - 2, 0);
		    result = (*itemPtr->typePtr->coordProc)(canvasPtr, itemPtr,
							    items - 2, argv);
		    Safefree(argv);
		    if (result != TCL_OK)
			croak(tclinterp->result);
		    
		    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
					 itemPtr->x2, itemPtr->y2);
		}
		else
		{
		    int argc;
		    char *nullp = NULL;
		    char **argv = &nullp;
		    double coord;

		    (*itemPtr->typePtr->coordProc)(canvasPtr,itemPtr, 0, argv);
		    /*
		     * The coordProc has just converted a list of doubles
		     * into ascii and formed a list of them in
		     * tclinterp->result. So now *we* split the list up
		     * again, convert each element back into a double and
		     * push them back on the perl stack. Double blech.
		     */
		    if (Tcl_SplitList(tclinterp, tclinterp->result,
				      &argc, &argv) != TCL_OK)
			croak("a canvas item coordproc returned rubbish");
		    for (; *argv; argv++)
		    {
			if (Tcl_GetDouble(tclinterp, *argv, &coord) != TCL_OK)
			    croak("a canvas item coordproc returned rubbish");
			av_store(stack, sp++, sv_2mortal(newSVnv(coord)));
		    }
		    if (argc > 0)
		    {
			sp--;
			RECOVERSTACK;
		    }
		}
	    }
	}
	break;

      case IX_CREATE:
	{
	    register Tk_ItemType *typePtr;
	    Tk_ItemType *matchPtr = NULL;
	    register Tk_Item *itemPtr;
	    char *itemname;
	    STRLEN length;
	    int result;
	    char **argv;
	    
	    if (items < 2)
		croak("wrong # args for create");

	    itemname = SvPV(st[2], length);
	    c = itemname[0];
	    for (typePtr = typeList; typePtr; typePtr = typePtr->nextPtr)
	    {
		if ((c == typePtr->name[0])
		    && (strncmp(itemname, typePtr->name, length) == 0))
		{
		    if (matchPtr)
			croak("ambiguous item type in create");
		    matchPtr = typePtr;
		}
	    }
	    if (!matchPtr)
		croak("bad item type in create");
	    
	    typePtr = matchPtr;
	    itemPtr = (Tk_Item *) ckalloc((unsigned) typePtr->itemSize);
	    itemPtr->id = canvasPtr->nextId;
	    canvasPtr->nextId++;
	    itemPtr->tagPtr = itemPtr->staticTagSpace;
	    itemPtr->tagSpace = TK_TAG_SPACE;
	    itemPtr->numTags = 0;
	    itemPtr->typePtr = typePtr;
	    argv = StackToArgv(sp + 3, items - 2, 0);
	    result = (*typePtr->createProc)(canvasPtr, itemPtr, items-2, argv);
	    Safefree(argv);
	    if (result != TCL_OK)
	    {
		ckfree((char *) itemPtr);
		croak("create failed");
	    }
	    itemPtr->nextPtr = NULL;
	    canvasPtr->hotPtr = itemPtr;
	    canvasPtr->hotPrevPtr = canvasPtr->lastItemPtr;
	    if (canvasPtr->lastItemPtr == NULL) {
		canvasPtr->firstItemPtr = itemPtr;
	    } else {
		canvasPtr->lastItemPtr->nextPtr = itemPtr;
	    }
	    canvasPtr->lastItemPtr = itemPtr;
	    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				 itemPtr->x2, itemPtr->y2);
	    canvasPtr->flags |= REPICK_NEEDED;
	    st[0] = sv_2mortal(newSViv(itemPtr->id));
	    break;
	}

      case IX_DCHARS:
	{
	    int result, first, last;
	    
	    if (items != 3 && items != 4)
		croak("wrong # args for dchars");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc == NULL)
		    || (itemPtr->typePtr->dCharsProc == NULL)) {
		    continue;
		}
		if ((*itemPtr->typePtr->indexProc)(canvasPtr, itemPtr,
			SvPV(st[3], na), &first) != TCL_OK)
		    croak("bad first index in dchars");
	    }
	    if (items == 4) {
		if ((*itemPtr->typePtr->indexProc)(canvasPtr, itemPtr,
			 SvPV(st[4], na), &last) != TCL_OK)
		    croak("bad last index in dchars");
	    } else {
		last = first;
	    }

	    /*
	     * Redraw both item's old and new areas:  it's possible
	     * that a delete could result in a new area larger than
	     * the old area.
	     */

	    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		    itemPtr->x2, itemPtr->y2);
	    result = (*itemPtr->typePtr->dCharsProc)(canvasPtr, itemPtr,
						     first, last);
	    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		    itemPtr->x2, itemPtr->y2);
	    if (result != TCL_OK)
		croak("dchars failed");
	    break;
	}

      case IX_DELETE:
	{
	    int i;

	    for (i = 2; i <= items; i++) {
		itemPtr = StartTagSearch(canvasPtr, SvPV(st[i], na), &search);
		for (; itemPtr != NULL; itemPtr = NextItem(&search)) {
		    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
					 itemPtr->x2, itemPtr->y2);
		    if (canvasPtr->bindingTable != NULL) {
			Tk_DeleteAllBindings(canvasPtr->bindingTable,
					     (ClientData) itemPtr);
		    }
		    (*itemPtr->typePtr->deleteProc)(canvasPtr, itemPtr);
		    if (itemPtr->tagPtr != itemPtr->staticTagSpace) {
			ckfree((char *) itemPtr->tagPtr);
		    }
		    if (search.prevPtr == NULL) {
			canvasPtr->firstItemPtr = itemPtr->nextPtr;
			if (canvasPtr->firstItemPtr == NULL) {
			    canvasPtr->lastItemPtr = NULL;
			}
		    } else {
			search.prevPtr->nextPtr = itemPtr->nextPtr;
		    }
		    if (canvasPtr->lastItemPtr == itemPtr) {
			canvasPtr->lastItemPtr = search.prevPtr;
		    }
		    ckfree((char *) itemPtr);
		    if (itemPtr == canvasPtr->currentItemPtr) {
			canvasPtr->currentItemPtr = NULL;
			canvasPtr->flags |= REPICK_NEEDED;
		    }
		    if (itemPtr == canvasPtr->focusItemPtr) {
			canvasPtr->focusItemPtr = NULL;
		    }
		    if (itemPtr == canvasPtr->selItemPtr) {
			canvasPtr->selItemPtr = NULL;
		    }
		    if ((itemPtr == canvasPtr->hotPtr)
			|| (itemPtr = canvasPtr->hotPrevPtr)) {
			canvasPtr->hotPtr = NULL;
		    }
		}
	    }
	    break;
	}

      case IX_DTAG:
	{
	    Tk_Uid tag;
	    int i;
	    char *tagorid;

	    if (items != 2 && items != 3)
		croak("wrong # args for dtag");

	    tagorid = SvPV(st[2], na);
	    tag = Tk_GetUid((items == 3) ? SvPV(st[3], na) : tagorid);

	    for (itemPtr = StartTagSearch(canvasPtr, tagorid, &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		for (i = itemPtr->numTags-1; i >= 0; i--) {
		    if (itemPtr->tagPtr[i] == tag) {
			itemPtr->tagPtr[i] = itemPtr->tagPtr[itemPtr->numTags-1];
			itemPtr->numTags--;
		    }
		}
	    }
	    break;
	}
	
      case IX_FIND:
	if (items < 2)
	    croak("wrong # args for find");

	sp = TkpFindItems(canvasPtr, sp, items, (char *) 0);
	break;
	
      case IX_FOCUS:
	if (items > 2)
	    croak("wrong # args for focus");

	itemPtr = canvasPtr->focusItemPtr;
	if (items == 1)
	{
	    if (itemPtr)
		st[0] = sv_2mortal(newSViv(itemPtr->id));
	    break;
	}
	if ((itemPtr != NULL) && (canvasPtr->flags & GOT_FOCUS)) {
	    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		    itemPtr->x2, itemPtr->y2);
	}
	if (st[2] == &sv_undef)
	{
	    canvasPtr->focusItemPtr = NULL;
	    break;
	}
	for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		itemPtr != NULL; itemPtr = NextItem(&search)) {
	    if (itemPtr->typePtr->icursorProc != NULL) {
		break;
	    }
	}
	if (!itemPtr)
	    break;

	canvasPtr->focusItemPtr = itemPtr;
	if (canvasPtr->flags & GOT_FOCUS) {
	    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		    itemPtr->x2, itemPtr->y2);
	}
	break;

      case IX_GETTAGS:
	if (items != 2)
	    croak("wrong # args for gettags");

	itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
	if (itemPtr != NULL && itemPtr->numTags > 0) {
	    int i;
	    for (i = 0; i < itemPtr->numTags; i++)
		av_store(stack, sp++,
			 sv_2mortal(newSVpv(itemPtr->tagPtr[i], 0)));
	    sp--;
	}
	break;
	
      case IX_ICURSOR:
	{
	    int index;

	    if (items != 3)
		croak("wrong # args for icursor");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc == NULL)
		    || (itemPtr->typePtr->icursorProc == NULL)) {
		    goto done;
		}
		if ((*itemPtr->typePtr->indexProc)(
		     canvasPtr, itemPtr, SvPV(st[3], na), &index) != TCL_OK) {
		    croak("index proc failed in icursor");
		    
		}
		(*itemPtr->typePtr->icursorProc)(canvasPtr, itemPtr, index);
		if ((itemPtr == canvasPtr->focusItemPtr)
		    && (canvasPtr->flags & CURSOR_ON)) {
		    EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
					 itemPtr->x2, itemPtr->y2);
		}
	    }
	    break;
	}
	
      case IX_INDEX:
	{
	    int index;

	    if (items != 3 )
		croak("wrong # args for tkindex");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if (itemPtr->typePtr->indexProc != NULL) {
		    break;
		}
	    }
	    if (itemPtr == NULL)
		croak("cannot find indexable item");

	    if ((*itemPtr->typePtr->indexProc)(
                canvasPtr, itemPtr, SvPV(st[3], na), &index) != TCL_OK)
		croak("index proc failed in tkindex");

	    st[0] = sv_2mortal(newSViv(index));
	    break;
	}
	
      case IX_INSERT:
	{
	    int beforeThis;
	    int result;
	    
	    if (items != 4)
		croak("wrong # args for insert");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc == NULL)
		    || (itemPtr->typePtr->insertProc == NULL)) {
		    continue;
		}
		if ((*itemPtr->typePtr->indexProc)(
                   canvasPtr, itemPtr, SvPV(st[3], na), &beforeThis) != TCL_OK)
		    croak("index proc failed in insert");

		/*
		 * Redraw both item's old and new areas:  it's possible
		 * that an insertion could result in a new area either
		 * larger or smaller than the old area.
		 */
		
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		result = (*itemPtr->typePtr->insertProc)(
                    canvasPtr, itemPtr, beforeThis, SvPV(st[4], na));
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		if (result != TCL_OK)
		    croak("insert proc failed in insert");
	    }
	    break;
	}
	
      case IX_ITEMCONFIGURE:
	if (items < 2)
	    croak("wrong # args for itemconfigure");

	for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
	     itemPtr != NULL; itemPtr = NextItem(&search)) {
	    if (items == 2)
		sp = TkpConfigureInfo(sp, canvasPtr->tkwin,
				      itemPtr->typePtr->configSpecs,
				      (char *) itemPtr, (char *) NULL, 0);
	    else if (items == 3)
		sp = TkpConfigureInfo(sp, canvasPtr->tkwin,
				      itemPtr->typePtr->configSpecs,
				      (char *) itemPtr, SvPV(st[3], na), 0);
	    else
	    {
		char **argv;
		int result;
		
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
			itemPtr->x2, itemPtr->y2);
		argv = StackToArgv(sp + 3, items - 2, 0);
		result = (*itemPtr->typePtr->configProc)(canvasPtr, itemPtr,
							 items - 2, argv,
							 TK_CONFIG_ARGV_ONLY);
		Safefree(argv);
		if (result != TCL_OK)
		    croak(tclinterp->result);
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		canvasPtr->flags |= REPICK_NEEDED;
	    }
	    if (items < 4)
		break;
	}
	break;
	
      case IX_LOWER:
	{
	    Tk_Item *prevPtr;

	    if (items != 2 && items != 3)
		croak("wrong # args for lower");

	    /*
	     * First find the item just after which we'll insert the
	     * named items.
	     */
	    
	    if (items == 2)
		prevPtr = NULL;
	    else
	    {
		prevPtr = StartTagSearch(canvasPtr, SvPV(st[3], na), &search);
		if (prevPtr)
		    prevPtr = search.prevPtr;
		else
		    croak("tag does not match any items in lower");
	    }
	    RelinkItems(canvasPtr, SvPV(st[2], na), prevPtr);
	    break;
	}
	
      case IX_MOVE:
	{
	    double xAmount, yAmount;
	    
	    if (items != 4)
		croak("wrong # args for move");
	    
	    if ((TkGetCanvasCoord(canvasPtr, SvPV(st[3], na),
				  &xAmount) != TCL_OK)
		|| (TkGetCanvasCoord(canvasPtr, SvPV(st[4], na),
				     &yAmount) != TCL_OK))
		croak("bad canvas coord(s) in move");
	    
	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		(void) (*itemPtr->typePtr->translateProc)(canvasPtr, itemPtr,
							  xAmount, yAmount);
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		canvasPtr->flags |= REPICK_NEEDED;
	    }
	    break;
	}
	
      case IX_POSTSCRIPT:
	{
	    char **argv;
	    int result;
	    
	    argv = StackToArgv(sp + 2, items - 1, 2);
	    result = TkCanvPostscriptCmd(canvasPtr, tclinterp, items-1, argv);
	    Safefree(argv);
	    if (result != TCL_OK)
		croak(tclinterp->result);
	    st[0] = sv_2mortal(newSVpv(tclinterp->result, 0));
	    break;
	}
	
      case IX_RAISE:
	{
	    Tk_Item *prevPtr;

	    if (items != 2 && items != 3)
		croak("wrong # args for raise");

	    /*
	     * First find the item just after which we'll insert the
	     * named items.
	     */
	    
	    if (items == 2)
		prevPtr = canvasPtr->lastItemPtr;
	    else
	    {
		prevPtr = NULL;
		itemPtr = StartTagSearch(canvasPtr, SvPV(st[3], na), &search);
		for (; itemPtr != NULL; itemPtr = NextItem(&search))
		    prevPtr = itemPtr;
		if (!prevPtr)
		    croak("tag does not match any items in raise");
	    }
	    RelinkItems(canvasPtr, SvPV(st[2], na), prevPtr);
	    break;
	}
	
      case IX_SCALE:
	{
	    double xOrigin, yOrigin, xScale, yScale;

	    if (items != 6)
		croak("wrong # args for scale");

	    if ((TkGetCanvasCoord(canvasPtr, SvPV(st[3], na),
				  &xOrigin) != TCL_OK)
		|| (TkGetCanvasCoord(canvasPtr, SvPV(st[4], na),
				     &yOrigin) != TCL_OK))
		croak("bad canvas coord(s) in scale");
		
	    xScale = SvNV(st[5]);
	    yScale = SvNV(st[6]);
	    
	    if ((xScale == 0.0) || (yScale == 0.0))
		croak("scale factors cannot be zero");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		(void) (*itemPtr->typePtr->scaleProc)(
                    canvasPtr, itemPtr, xOrigin, yOrigin, xScale, yScale);
		EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
				     itemPtr->x2, itemPtr->y2);
		canvasPtr->flags |= REPICK_NEEDED;
	    }
	    break;
	}
	
      case IX_SCANMARK:
	if (items != 3)
	    croak("wrong # args for scanmark");
	
	canvasPtr->scanX = SvIV(st[2]);
	canvasPtr->scanXOrigin = canvasPtr->xOrigin;
	canvasPtr->scanY = SvIV(st[3]);
	canvasPtr->scanYOrigin = canvasPtr->yOrigin;
	break;

      case IX_SCANDRAGTO:
	{
	    int newXOrigin, newYOrigin, tmp;

	    /*
	     * Compute a new view origin for the canvas, amplifying the
	     * mouse motion and rounding to the nearest multiple of the
	     * scroll increment.
	     */

	    tmp = canvasPtr->scanXOrigin - 10*(SvIV(st[2]) - canvasPtr->scanX)
		- canvasPtr->scrollX1;
	    if (tmp >= 0) {
		tmp = (tmp + canvasPtr->scrollIncrement/2)
		    /canvasPtr->scrollIncrement;
	    } else {
		tmp = -(((-tmp) + canvasPtr->scrollIncrement/2)
			/canvasPtr->scrollIncrement);
	    }
	    newXOrigin = canvasPtr->scrollX1 + tmp*canvasPtr->scrollIncrement;
	    tmp = canvasPtr->scanYOrigin - 10*(SvIV(st[3]) - canvasPtr->scanY)
		- canvasPtr->scrollY1;
	    if (tmp >= 0) {
		tmp = (tmp + canvasPtr->scrollIncrement/2)
		    /canvasPtr->scrollIncrement;
	    } else {
		tmp = -(((-tmp) + canvasPtr->scrollIncrement/2)
			/canvasPtr->scrollIncrement);
	    }
	    newYOrigin = canvasPtr->scrollY1 + tmp*canvasPtr->scrollIncrement;
	    CanvasSetOrigin(canvasPtr, newXOrigin, newYOrigin);
	    break;
	}

      case IX_SELECTADJUST:
	{
	    int index;

	    if (items != 3)
		croak("wrong # args in selectadjust");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc != NULL)
		    && (itemPtr->typePtr->selectionProc != NULL)){
		    break;
		}
	    }
	    if (!itemPtr)
		croak("cannot find indexable and selectable item");

	    if ((*itemPtr->typePtr->indexProc)(
	        canvasPtr, itemPtr, SvPV(st[3], na), &index) != TCL_OK)
		croak("index proc failed in tkselect");
	
	    if (canvasPtr->selItemPtr == itemPtr) {
		if (index < (canvasPtr->selectFirst
			+ canvasPtr->selectLast)/2) {
		    canvasPtr->selectAnchor = canvasPtr->selectLast + 1;
		} else {
		    canvasPtr->selectAnchor = canvasPtr->selectFirst;
		}
	    }
	    CanvasSelectTo(canvasPtr, itemPtr, index);
	    break;
	}
	
      case IX_SELECTCLEAR:
	if (items != 1)
	    croak("wrong # args for selectclear");

	if (canvasPtr->selItemPtr != NULL) {
	    EventuallyRedrawArea(canvasPtr, canvasPtr->selItemPtr->x1,
		canvasPtr->selItemPtr->y1, canvasPtr->selItemPtr->x2,
		canvasPtr->selItemPtr->y2);
	    canvasPtr->selItemPtr = NULL;
	}
	break;
	
      case IX_SELECTFROM:
	{
	    int index;

	    if (items != 3)
		croak("wrong # args in selectfrom");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc != NULL)
		    && (itemPtr->typePtr->selectionProc != NULL)){
		    break;
		}
	    }
	    if (!itemPtr)
		croak("cannot find indexable and selectable item");

	    if ((*itemPtr->typePtr->indexProc)(
	        canvasPtr, itemPtr, SvPV(st[3], na), &index) != TCL_OK)
		croak("index proc failed in tkselect");

	    canvasPtr->anchorItemPtr = itemPtr;
	    canvasPtr->selectAnchor = index;
	    break;
	}
	
      case IX_SELECTITEM:
	if (items != 1)
	    croak("wrong # args for selectitem");

	if (canvasPtr->selItemPtr)
	    st[0] = sv_2mortal(newSViv(canvasPtr->selItemPtr->id));
	break;
	
      case IX_SELECTTO:
	{
	    int index;

	    if (items != 3)
		croak("wrong # args in selectto");

	    for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		 itemPtr != NULL; itemPtr = NextItem(&search)) {
		if ((itemPtr->typePtr->indexProc != NULL)
		    && (itemPtr->typePtr->selectionProc != NULL)){
		    break;
		}
	    }
	    if (!itemPtr)
		croak("cannot find indexable and selectable item");

	    if ((*itemPtr->typePtr->indexProc)(
	        canvasPtr, itemPtr, SvPV(st[3], na), &index) != TCL_OK)
		croak("index proc failed in tkselect");

	    CanvasSelectTo(canvasPtr, itemPtr, index);
	    break;
	}

      case IX_TYPE:
	if (items != 2)
	    croak("wrong # args for type");
	
	itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
	if (itemPtr)
	    st[0] = sv_2mortal(newSVpv(itemPtr->typePtr->name, 0));
	break;
	
      case IX_XVIEW:
	if (items != 2)
	    croak("wrong # args for xview");
	
	CanvasSetOrigin(canvasPtr,
	    (canvasPtr->scrollX1 + SvIV(st[2])*canvasPtr->scrollIncrement),
	    canvasPtr->yOrigin);
	break;

      case IX_YVIEW:
	if (items != 2)
	    croak("wrong # args for yview");
	
	CanvasSetOrigin(canvasPtr, canvasPtr->xOrigin,
	    (canvasPtr->scrollY1 + SvIV(st[2])*canvasPtr->scrollIncrement));
	break;
    }
    
    done:
    Tk_Release((ClientData) canvasPtr);
    return sp;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyCanvas --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a canvas at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the canvas is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyCanvas(clientData)
    ClientData clientData;	/* Info about canvas widget. */
{
    register Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;
    register Tk_Item *itemPtr;

    /*
     * Free up all of the items in the canvas.
     */

    for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	    itemPtr = canvasPtr->firstItemPtr) {
	canvasPtr->firstItemPtr = itemPtr->nextPtr;
	(*itemPtr->typePtr->deleteProc)(canvasPtr, itemPtr);
	if (itemPtr->tagPtr != itemPtr->staticTagSpace) {
	    ckfree((char *) itemPtr->tagPtr);
	}
	ckfree((char *) itemPtr);
    }

    /*
     * Free up all the stuff that requires special handling,
     * then let Tk_FreeOptions handle all the standard option-related
     * stuff.
     */

    if (canvasPtr->pixmapGC != None) {
	Tk_FreeGC(canvasPtr->display, canvasPtr->pixmapGC);
    }
    Tk_DeleteTimerHandler(canvasPtr->insertBlinkHandler);
    if (canvasPtr->bindingTable != NULL) {
	Tk_DeleteBindingTable(canvasPtr->bindingTable);
    }
    Tk_FreeOptions(configSpecs, (char *) canvasPtr, canvasPtr->display, 0);
    ckfree((char *) canvasPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpConfigureCanvas --
 *
 *	This procedure is called to process the perl stack, plus
 *	the Tk option database, in order to configure (or
 *	reconfigure) a canvas widget.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for canvasPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static void
TkpConfigureCanvas(canvasPtr, sp, items, flags)
    register Tk_Canvas *canvasPtr;	/* Information about widget;  may or may
				 * not already have values for some fields. */
    int sp;
    int items; 
    int flags;			/* Flags to pass to TkpConfigureWidget. */
{
    XGCValues gcValues;
    GC new;

    TkpConfigureWidget(canvasPtr->tkwin, configSpecs,
		       sp, items, (char *) canvasPtr, flags);

    /*
     * A few options need special processing, such as setting the
     * background from a 3-D border and creating a GC for copying
     * bits to the screen.
     */

    Tk_SetBackgroundFromBorder(canvasPtr->tkwin, canvasPtr->bgBorder);

    gcValues.function = GXcopy;
    gcValues.foreground = Tk_3DBorderColor(canvasPtr->bgBorder)->pixel;
    gcValues.graphics_exposures = False;
    new = Tk_GetGC(canvasPtr->tkwin,
	    GCFunction|GCForeground|GCGraphicsExposures, &gcValues);
    if (canvasPtr->pixmapGC != None) {
	Tk_FreeGC(canvasPtr->display, canvasPtr->pixmapGC);
    }
    canvasPtr->pixmapGC = new;

    /*
     * Reset the desired dimensions for the window.
     */

    Tk_GeometryRequest(canvasPtr->tkwin, canvasPtr->width, canvasPtr->height);

    /*
     * Restart the cursor timing sequence in case the on-time or off-time
     * just changed.
     */

    if (canvasPtr->flags & GOT_FOCUS) {
	CanvasFocusProc(canvasPtr, 1);
    }

    /*
     * Recompute the scroll region.
     */

    canvasPtr->scrollX1 = 0;
    canvasPtr->scrollY1 = 0;
    canvasPtr->scrollX2 = 0;
    canvasPtr->scrollY2 = 0;
    if (canvasPtr->regionString != NULL) {
	int argc2;
	char **argv2;

	if (Tcl_SplitList(tclinterp, canvasPtr->regionString,
			  &argc2, &argv2) != TCL_OK
	    || argc2 != 4) {
	    ckfree(canvasPtr->regionString);
	    ckfree((char *) argv2);
	    canvasPtr->regionString = NULL;
	    croak("bad scroll region");
	}
	if ((Tk_GetPixels(tclinterp, canvasPtr->tkwin,
		    argv2[0], &canvasPtr->scrollX1) != TCL_OK)
		|| (Tk_GetPixels(tclinterp, canvasPtr->tkwin,
		    argv2[1], &canvasPtr->scrollY1) != TCL_OK)
		|| (Tk_GetPixels(tclinterp, canvasPtr->tkwin,
		    argv2[2], &canvasPtr->scrollX2) != TCL_OK)
		|| (Tk_GetPixels(tclinterp, canvasPtr->tkwin,
		    argv2[3], &canvasPtr->scrollY2) != TCL_OK)) {
	    ckfree(canvasPtr->regionString);
	    ckfree((char *) argv2);
	    canvasPtr->regionString = NULL;
	    croak("bad scroll region");
	}
	ckfree((char *) argv2);
    }

    /*
     * Reset the canvases origin (this is a no-op unless confine
     * mode has just been turned on or the scroll region has changed).
     */

    CanvasSetOrigin(canvasPtr, canvasPtr->xOrigin, canvasPtr->yOrigin);
    canvasPtr->flags |= UPDATE_SCROLLBARS;
    EventuallyRedrawArea(canvasPtr, canvasPtr->xOrigin, canvasPtr->yOrigin,
	    canvasPtr->xOrigin + Tk_Width(canvasPtr->tkwin),
	    canvasPtr->yOrigin + Tk_Height(canvasPtr->tkwin));
}

/*
 *--------------------------------------------------------------
 *
 * DisplayCanvas --
 *
 *	This procedure redraws the contents of a canvas window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
DisplayCanvas(clientData)
    ClientData clientData;	/* Information about widget. */
{
    register Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;
    register Tk_Window tkwin = canvasPtr->tkwin;
    register Tk_Item *itemPtr;
    Pixmap pixmap;
    int screenX1, screenX2, screenY1, screenY2;

    if (canvasPtr->tkwin == NULL) {
	return;
    }
    if (!Tk_IsMapped(tkwin)) {
	goto done;
    }

    /*
     * Choose a new current item if that is needed (this could cause
     * event handlers to be invoked).
     */

    while (canvasPtr->flags & REPICK_NEEDED) {
	Tk_Preserve((ClientData) canvasPtr);
	canvasPtr->flags &= ~REPICK_NEEDED;
	PickCurrentItem(canvasPtr, &canvasPtr->pickEvent);
	tkwin = canvasPtr->tkwin;
	Tk_Release((ClientData) canvasPtr);
	if (tkwin == NULL) {
	    return;
	}
    }

    /*
     * Compute the intersection between the area that needs redrawing
     * and the area that's visible on the screen.
     */

    screenX1 = canvasPtr->xOrigin;
    screenY1 = canvasPtr->yOrigin;
    screenX2 = screenX1 + Tk_Width(tkwin);
    screenY2 = screenY1 + Tk_Height(tkwin);
    if (canvasPtr->redrawX1 > screenX1) {
	screenX1 = canvasPtr->redrawX1;
    }
    if (canvasPtr->redrawY1 > screenY1) {
	screenY1 = canvasPtr->redrawY1;
    }
    if (canvasPtr->redrawX2 < screenX2) {
	screenX2 = canvasPtr->redrawX2;
    }
    if (canvasPtr->redrawY2 < screenY2) {
	screenY2 = canvasPtr->redrawY2;
    }
    if ((screenX1 >= screenX2) || (screenY1 >= screenY2)) {
	goto done;
    }

    /*
     * Redrawing is done in a temporary pixmap that is allocated
     * here and freed at the end of the procedure.  All drawing
     * is done to the pixmap, and the pixmap is copied to the
     * screen at the end of the procedure. The temporary pixmap
     * serves two purposes:
     *
     * 1. It provides a smoother visual effect (no clearing and
     *    gradual redraw will be visible to users).
     * 2. It allows us to redraw only the objects that overlap
     *    the redraw area.  Otherwise incorrect results could
     *	  occur from redrawing things that stick outside of
     *	  the redraw area (we'd have to redraw everything in
     *    order to make the overlaps look right).
     *
     * Some tricky points about the pixmap:
     *
     * 1. We only allocate a large enough pixmap to hold the
     *    area that has to be redisplayed.  This saves time in
     *    in the X server for large objects that cover much
     *    more than the area being redisplayed:  only the area
     *    of the pixmap will actually have to be redrawn.
     * 2. Some X servers (e.g. the one for DECstations) have troubles
     *    with characters that overlap an edge of the pixmap (on the
     *    DEC servers, as of 8/18/92, such characters are drawn one
     *    pixel too far to the right).  To handle this problem,
     *    make the pixmap a bit larger than is absolutely needed
     *    so that for normal-sized fonts the characters that overlap
     *    the edge of the pixmap will be outside the area we care
     *    about.
     */

    canvasPtr->drawableXOrigin = screenX1 - 30;
    canvasPtr->drawableYOrigin = screenY1 - 30;
    pixmap = XCreatePixmap(Tk_Display(tkwin), Tk_WindowId(tkwin),
	screenX2 + 30 - canvasPtr->drawableXOrigin,
	screenY2 + 30 - canvasPtr->drawableYOrigin,
	Tk_Depth(tkwin));

    /*
     * Clear the area to be redrawn.
     */

    XFillRectangle(Tk_Display(tkwin), pixmap, canvasPtr->pixmapGC,
	    screenX1 - canvasPtr->drawableXOrigin,
	    screenY1 - canvasPtr->drawableYOrigin,
	    (unsigned int) (screenX2 - screenX1),
	    (unsigned int) (screenY2 - screenY1));

    /*
     * Scan through the item list, redrawing those items that need it.
     * An item must be redraw if either (a) it intersects the smaller
     * on-screen area or (b) it intersects the full canvas area and its
     * type requests that it be redrawn always (e.g. so subwindows can
     * be unmapped when they move off-screen).
     */

    for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	    itemPtr = itemPtr->nextPtr) {
	if ((itemPtr->x1 >= screenX2)
		|| (itemPtr->y1 >= screenY2)
		|| (itemPtr->x2 < screenX1)
		|| (itemPtr->y2 < screenY1)) {
	    if (!itemPtr->typePtr->alwaysRedraw
		    || (itemPtr->x1 >= canvasPtr->redrawX2)
		    || (itemPtr->y1 >= canvasPtr->redrawY2)
		    || (itemPtr->x2 < canvasPtr->redrawX1)
		    || (itemPtr->y2 < canvasPtr->redrawY1)) {
		continue;
	    }
	}
	(*itemPtr->typePtr->displayProc)(canvasPtr, itemPtr, pixmap);
    }

    /*
     * Draw the window border.
     */

    if (canvasPtr->relief != TK_RELIEF_FLAT) {
	Tk_Draw3DRectangle(Tk_Display(tkwin), pixmap,
		canvasPtr->bgBorder,
		canvasPtr->xOrigin - canvasPtr->drawableXOrigin,
		canvasPtr->yOrigin - canvasPtr->drawableYOrigin,
		Tk_Width(tkwin), Tk_Height(tkwin),
		canvasPtr->borderWidth, canvasPtr->relief);
    }

    /*
     * Copy from the temporary pixmap to the screen, then free up
     * the temporary pixmap.
     */

    XCopyArea(Tk_Display(tkwin), pixmap, Tk_WindowId(tkwin),
	    canvasPtr->pixmapGC,
	    screenX1 - canvasPtr->drawableXOrigin,
	    screenY1 - canvasPtr->drawableYOrigin,
	    screenX2 - screenX1, screenY2 - screenY1,
	    screenX1 - canvasPtr->xOrigin, screenY1 - canvasPtr->yOrigin);
    XFreePixmap(Tk_Display(tkwin), pixmap);

    done:
    canvasPtr->flags &= ~REDRAW_PENDING;
    if (canvasPtr->flags & UPDATE_SCROLLBARS) {
	CanvasUpdateScrollbars(canvasPtr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * CanvasEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on canvases.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
CanvasEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;

    if (eventPtr->type == Expose) {
	int x, y;

	x = eventPtr->xexpose.x + canvasPtr->xOrigin;
	y = eventPtr->xexpose.y + canvasPtr->yOrigin;
	EventuallyRedrawArea(canvasPtr, x, y, x + eventPtr->xexpose.width,
		y + eventPtr->xexpose.height);
    } else if (eventPtr->type == DestroyNotify) {
	TkpDeleteWidgetHash(canvasPtr->tkwin);
	canvasPtr->tkwin = NULL;
	if (canvasPtr->flags & REDRAW_PENDING) {
	    Tk_CancelIdleCall(DisplayCanvas, (ClientData) canvasPtr);
	}
	Tk_EventuallyFree((ClientData) canvasPtr, DestroyCanvas);
    } else if (eventPtr->type == ConfigureNotify) {
	canvasPtr->flags |= UPDATE_SCROLLBARS;

	/*
	 * The call below is needed in order to recenter the canvas if
	 * it's confined and its scroll region is smaller than the window.
	 */

	CanvasSetOrigin(canvasPtr, canvasPtr->xOrigin, canvasPtr->yOrigin);
	EventuallyRedrawArea(canvasPtr, 0, 0, Tk_Width(canvasPtr->tkwin),
		Tk_Height(canvasPtr->tkwin));
    } else if (eventPtr->type == FocusIn) {
	CanvasFocusProc(canvasPtr, 1);
    } else if (eventPtr->type == FocusOut) {
	CanvasFocusProc(canvasPtr, 0);
    } else if (eventPtr->type == UnmapNotify) {
	register Tk_Item *itemPtr;
	
	/*
	 * Special hack:  if the canvas is unmapped, then must notify
	 * all items with "alwaysRedraw" set, so that they know that
	 * they are no longer displayed.
	 */
	
	for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextPtr) {
	    if (itemPtr->typePtr->alwaysRedraw) {
		(*itemPtr->typePtr->displayProc)(canvasPtr, itemPtr, None);
	    }
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * EventuallyRedrawArea --
 *
 *	Arrange for part or all of a canvas widget to redrawn at
 *	the next convenient time in the future.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The screen will eventually be refreshed.
 *
 *--------------------------------------------------------------
 */

static void
EventuallyRedrawArea(canvasPtr, x1, y1, x2, y2)
    register Tk_Canvas *canvasPtr;	/* Information about widget. */
    int x1, y1;				/* Upper left corner of area to
					 * redraw.  Pixels on edge are
					 * redrawn. */
    int x2, y2;				/* Lower right corner of area to
					 * redraw.  Pixels on edge are
					 * not redrawn. */
{
    if ((canvasPtr->tkwin == NULL) || !Tk_IsMapped(canvasPtr->tkwin)) {
	return;
    }
    if (canvasPtr->flags & REDRAW_PENDING) {
	if (x1 <= canvasPtr->redrawX1) {
	    canvasPtr->redrawX1 = x1;
	}
	if (y1 <= canvasPtr->redrawY1) {
	    canvasPtr->redrawY1 = y1;
	}
	if (x2 >= canvasPtr->redrawX2) {
	    canvasPtr->redrawX2 = x2;
	}
	if (y2 >= canvasPtr->redrawY2) {
	    canvasPtr->redrawY2 = y2;
	}
    } else {
	canvasPtr->redrawX1 = x1;
	canvasPtr->redrawY1 = y1;
	canvasPtr->redrawX2 = x2;
	canvasPtr->redrawY2 = y2;
	Tk_DoWhenIdle(DisplayCanvas, (ClientData) canvasPtr);
	canvasPtr->flags |= REDRAW_PENDING;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Tk_CreateItemType --
 *
 *	This procedure may be invoked to add a new kind of canvas
 *	element to the core item types supported by Tk.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	From now on, the new item type will be useable in canvas
 *	widgets (e.g. typePtr->name can be used as the item type
 *	in "create" widget commands).  If there was already a
 *	type with the same name as in typePtr, it is replaced with
 *	the new type.
 *
 *--------------------------------------------------------------
 */

void
Tk_CreateItemType(typePtr)
    Tk_ItemType *typePtr;		/* Information about item type;
					 * storage must be statically
					 * allocated (must live forever). */
{
    if (typeList == NULL) {
	InitCanvas();
    }
    typePtr->nextPtr = typeList;
    typeList = typePtr;
}

/*
 *--------------------------------------------------------------
 *
 * InitCanvas --
 *
 *	This procedure is invoked to perform once-only-ever
 *	initialization for the module, such as setting up
 *	the type table.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static void
InitCanvas()
{
    if (typeList != NULL) {
	return;
    }
    typeList = &TkRectangleType;
    TkRectangleType.nextPtr = &TkTextType;
    TkTextType.nextPtr = &TkPolygonType;
    TkPolygonType.nextPtr = &TkOvalType;
    TkOvalType.nextPtr = &TkLineType;
    TkLineType.nextPtr = &TkWindowType;
    TkWindowType.nextPtr = &TkBitmapType;
    TkBitmapType.nextPtr = &TkArcType;
    TkArcType.nextPtr = NULL;
    allUid = Tk_GetUid("all");
    currentUid = Tk_GetUid("current");
}

/*
 *--------------------------------------------------------------
 *
 * StartTagSearch --
 *
 *	This procedure is called to initiate an enumeration of
 *	all items in a given canvas that contain a given tag.
 *
 * Results:
 *	The return value is a pointer to the first item in
 *	canvasPtr that matches tag, or NULL if there is no
 *	such item.  The information at *searchPtr is initialized
 *	such that successive calls to NextItem will return
 *	successive items that match tag.
 *
 * Side effects:
 *	SearchPtr is linked into a list of searches in progress
 *	on canvasPtr, so that elements can safely be deleted
 *	while the search is in progress.  EndTagSearch must be
 *	called at the end of the search to unlink searchPtr from
 *	this list.
 *
 *--------------------------------------------------------------
 */

static Tk_Item *
StartTagSearch(canvasPtr, tag, searchPtr)
    Tk_Canvas *canvasPtr;		/* Canvas whose items are to be
					 * searched. */
    char *tag;				/* String giving tag value. */
    TagSearch *searchPtr;		/* Record describing tag search;
					 * will be initialized here. */
{
    int id;
    register Tk_Item *itemPtr, *prevPtr;
    register Tk_Uid *tagPtr;
    register Tk_Uid uid;
    register int count;

    /*
     * Initialize the search.
     */

    searchPtr->canvasPtr = canvasPtr;
    searchPtr->searchOver = 0;

    /*
     * Find the first matching item in one of several ways. If the tag
     * is a number then it selects the single item with the matching
     * identifier.  In this case see if the item being requested is the
     * hot item, in which case the search can be skipped.
     */

    if (isdigit(UCHAR(*tag))) {
	char *end;

	numIdSearches++;
	id = strtoul(tag, &end, 0);
	if (*end == 0) {
	    itemPtr = canvasPtr->hotPtr;
	    prevPtr = canvasPtr->hotPrevPtr;
	    if ((itemPtr == NULL) || (itemPtr->id != id) || (prevPtr == NULL)
		    || (prevPtr->nextPtr != itemPtr)) {
		numSlowSearches++;
		for (prevPtr = NULL, itemPtr = canvasPtr->firstItemPtr;
			itemPtr != NULL;
			prevPtr = itemPtr, itemPtr = itemPtr->nextPtr) {
		    if (itemPtr->id == id) {
			break;
		    }
		}
	    }
	    searchPtr->prevPtr = prevPtr;
	    searchPtr->searchOver = 1;
	    canvasPtr->hotPtr = itemPtr;
	    canvasPtr->hotPrevPtr = prevPtr;
	    return itemPtr;
	}
    }

    searchPtr->tag = uid = Tk_GetUid(tag);
    if (uid == allUid) {

	/*
	 * All items match.
	 */

	searchPtr->tag = NULL;
	searchPtr->prevPtr = NULL;
	searchPtr->currentPtr = canvasPtr->firstItemPtr;
	return canvasPtr->firstItemPtr;
    }

    /*
     * None of the above.  Search for an item with a matching tag.
     */

    for (prevPtr = NULL, itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	    prevPtr = itemPtr, itemPtr = itemPtr->nextPtr) {
	for (tagPtr = itemPtr->tagPtr, count = itemPtr->numTags;
		count > 0; tagPtr++, count--) {
	    if (*tagPtr == uid) {
		searchPtr->prevPtr = prevPtr;
		searchPtr->currentPtr = itemPtr;
		return itemPtr;
	    }
	}
    }
    searchPtr->prevPtr = prevPtr;
    searchPtr->searchOver = 1;
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * NextItem --
 *
 *	This procedure returns successive items that match a given
 *	tag;  it should be called only after StartTagSearch has been
 *	used to begin a search.
 *
 * Results:
 *	The return value is a pointer to the next item that matches
 *	the tag specified to StartTagSearch, or NULL if no such
 *	item exists.  *SearchPtr is updated so that the next call
 *	to this procedure will return the next item.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static Tk_Item *
NextItem(searchPtr)
    TagSearch *searchPtr;		/* Record describing search in
					 * progress. */
{
    register Tk_Item *itemPtr, *prevPtr;
    register int count;
    register Tk_Uid uid;
    register Tk_Uid *tagPtr;

    /*
     * Find next item in list (this may not actually be a suitable
     * one to return), and return if there are no items left.
     */

    prevPtr = searchPtr->prevPtr;
    if (prevPtr == NULL) {
	itemPtr = searchPtr->canvasPtr->firstItemPtr;
    } else {
	itemPtr = prevPtr->nextPtr;
    }
    if ((itemPtr == NULL) || (searchPtr->searchOver)) {
	searchPtr->searchOver = 1;
	return NULL;
    }
    if (itemPtr != searchPtr->currentPtr) {
	/*
	 * The structure of the list has changed.  Probably the
	 * previously-returned item was removed from the list.
	 * In this case, don't advance prevPtr;  just return
	 * its new successor (i.e. do nothing here).
	 */
    } else {
	prevPtr = itemPtr;
	itemPtr = prevPtr->nextPtr;
    }

    /*
     * Handle special case of "all" search by returning next item.
     */

    uid = searchPtr->tag;
    if (uid == NULL) {
	searchPtr->prevPtr = prevPtr;
	searchPtr->currentPtr = itemPtr;
	return itemPtr;
    }

    /*
     * Look for an item with a particular tag.
     */

    for ( ; itemPtr != NULL; prevPtr = itemPtr, itemPtr = itemPtr->nextPtr) {
	for (tagPtr = itemPtr->tagPtr, count = itemPtr->numTags;
		count > 0; tagPtr++, count--) {
	    if (*tagPtr == uid) {
		searchPtr->prevPtr = prevPtr;
		searchPtr->currentPtr = itemPtr;
		return itemPtr;
	    }
	}
    }
    searchPtr->prevPtr = prevPtr;
    searchPtr->searchOver = 1;
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * TkpAddTag --
 *
 *	This is a utility procedure called by TkpFindItems.  It
 *	adds a new tag to itemPtr.
 *
 * Side effects:
 *	tag is added to itemPtr's list of tags.
 *
 *--------------------------------------------------------------
 */

static void
TkpAddTag(itemPtr, tag)
register Tk_Item *itemPtr;		/* Item to modify. */
Tk_Uid tag;				/* Tag to add to those already
					 * present for item. */
{
    register Tk_Uid *tagPtr;
    register int count;

    for (tagPtr = itemPtr->tagPtr, count = itemPtr->numTags;
	    count > 0; tagPtr++, count--) {
	if (tag == *tagPtr)
	    return;
    }

    /*
     * Grow the tag space if there's no more room left in the current
     * block.
     */

    if (itemPtr->tagSpace == itemPtr->numTags) {
	Tk_Uid *newTagPtr;
	
	itemPtr->tagSpace += 5;
	newTagPtr = (Tk_Uid *) ckalloc((unsigned)
		(itemPtr->tagSpace * sizeof(Tk_Uid)));
	memcpy((VOID *) newTagPtr, (VOID *) itemPtr->tagPtr,
		(itemPtr->numTags * sizeof(Tk_Uid)));
	if (itemPtr->tagPtr != itemPtr->staticTagSpace) {
	    ckfree((char *) itemPtr->tagPtr);
	}
	itemPtr->tagPtr = newTagPtr;
	tagPtr = &itemPtr->tagPtr[itemPtr->numTags];
    }
    
    /*
     * Add in the new tag.
     */
    
    *tagPtr = tag;
    itemPtr->numTags++;
    return;
}

/*
 *--------------------------------------------------------------
 *
 * TkpFindItems --
 *
 *	This procedure does all the work of implementing the
 *	"find" and "addtag" options of the canvas widget command,
 *	which locate items that have certain features (location,
 *	tags, position in display list, etc.).
 *
 * Side effects:
 *	If newTag is non-NULL, then all the items that match the
 *	information on the stack have that tag added to their
 *	lists of tags.
 *
 *--------------------------------------------------------------
 */

static int
TkpFindItems(canvasPtr, sp, items, newTag)
    Tk_Canvas *canvasPtr;		/* Canvas whose items are to be
					 * searched. */
    int sp;
    int items;
    char *newTag;
{
    SV **st = stack_base + sp;
    char c;
    STRLEN length;
    TagSearch search;
    register Tk_Item *itemPtr;
    Tk_Uid uid;
    char *searchcom;
    
    uid = newTag ? Tk_GetUid(newTag) : NULL;
    searchcom = SvPV(st[2], length);
    c = searchcom[0];

    if ((c == 'a') && (strncmp(searchcom, "above", length) == 0)
	    && (length >= 2)) {
	Tk_Item *lastPtr = NULL;
	if (items != 3)
	    croak("wrong # args for searchspec above");
	for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[3], na), &search);
	     itemPtr != NULL; itemPtr = NextItem(&search)) {
	    lastPtr = itemPtr;
	}
	if ((lastPtr != NULL) && (lastPtr->nextPtr != NULL)) {
	    if (uid)
		TkpAddTag(lastPtr->nextPtr, uid);
	    else
		st[0] = sv_2mortal(newSViv(lastPtr->nextPtr->id));
	}
    } else if ((c == 'a') && (strncmp(searchcom, "all", length) == 0)
	    && (length >= 2)) {
	int pushedany = 0;

	if (items != 2)
	    croak("wrong # args for searchspec all");
	for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextPtr) {
	    if (uid)
		TkpAddTag(itemPtr, uid);
	    else
	    {
		av_store(stack, sp++, sv_2mortal(newSViv(itemPtr->id)));
		pushedany++;
	    }
	}
	if (pushedany)
	{
	    sp--;
	    RECOVERSTACK;
	}
    } else if ((c == 'b') && (strncmp(searchcom, "below", length) == 0)) {
	if (items != 3)
	    croak("wrong # args for searchspec below");
	itemPtr = StartTagSearch(canvasPtr, SvPV(st[3], na), &search);
	if (search.prevPtr != NULL) {
	    if (uid)
		TkpAddTag(search.prevPtr, uid);
	    else
		st[0] = sv_2mortal(newSViv(search.prevPtr->id));
	}
    } else if ((c == 'c') && (strncmp(searchcom, "closest", length) == 0)) {
	double closestDist;
	Tk_Item *startPtr, *closestPtr;
	double coords[2], halo;
	int x1, y1, x2, y2;

	if (items < 4 || items > 6)
	    croak("wrong # args for searchspec closest");
	if ((TkGetCanvasCoord(canvasPtr, SvPV(st[3], na), &coords[0]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr, SvPV(st[4], na), &coords[1])
		!= TCL_OK))
	    croak("bad canvas coord in searshspec closest");
	
	if (items > 4)
	{
	    if (TkGetCanvasCoord(canvasPtr, SvPV(st[5], na), &halo) != TCL_OK)
		croak("bad canvas coord in searshspec closest");
	    if (halo < 0.0)
		croak("cannot have negative halo value in searchspec closest");
	}
	else
	    halo = 0.0;
	
	/*
	 * Find the item at which to start the search.
	 */

	startPtr = canvasPtr->firstItemPtr;
	if (items == 6) {
	    itemPtr = StartTagSearch(canvasPtr, SvPV(st[6], na), &search);
	    if (itemPtr != NULL) {
		startPtr = itemPtr;
	    }
	}

	/*
	 * The code below is optimized so that it can eliminate most
	 * items without having to call their item-specific procedures.
	 * This is done by keeping a bounding box (x1, y1, x2, y2) that
	 * an item's bbox must overlap if the item is to have any
	 * chance of being closer than the closest so far.
	 */

	itemPtr = startPtr;
	if (itemPtr == NULL) {
	    return sp;
	}
	closestDist = (*itemPtr->typePtr->pointProc)(canvasPtr,
		itemPtr, coords) - halo;
	if (closestDist < 0.0) {
	    closestDist = 0.0;
	}
	while (1) {
	    double newDist;
	    
	    /*
	     * Update the bounding box using itemPtr, which is the
	     * new closest item.
	     */

	    x1 = (coords[0] - closestDist - halo - 1);
	    y1 = (coords[1] - closestDist - halo - 1);
	    x2 = (coords[0] + closestDist + halo + 1);
	    y2 = (coords[1] + closestDist + halo + 1);
	    closestPtr = itemPtr;

	    /*
	     * Search for an item that beats the current closest one.
	     * Work circularly through the canvas's item list until
	     * getting back to the starting item.
	     */

	    while (1) {
		itemPtr = itemPtr->nextPtr;
		if (itemPtr == NULL) {
		    itemPtr = canvasPtr->firstItemPtr;
		}
		if (itemPtr == startPtr) {
		    if (uid)
			TkpAddTag(closestPtr, uid);
		    else
			st[0] = sv_2mortal(newSViv(closestPtr->id));
		    return sp;
		}
		if ((itemPtr->x1 >= x2) || (itemPtr->x2 <= x1)
			|| (itemPtr->y1 >= y2) || (itemPtr->y2 <= y1)) {
		    continue;
		}
		newDist = (*itemPtr->typePtr->pointProc)(canvasPtr,
			itemPtr, coords) - halo;
		if (newDist < 0.0) {
		    newDist = 0.0;
		}
		if (newDist <= closestDist) {
		    closestDist = newDist;
		    break;
		}
	    }
	}
    } else if ((c == 'e') && (strncmp(searchcom, "enclosed", length) == 0)) {
	double rect[4];
	
	if (items != 6)
	    croak("wrong # args for searchspec enclosed");
	    
	if ((TkGetCanvasCoord(canvasPtr, SvPV(st[3], na), &rect[0]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[4],na),&rect[1]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[5],na),&rect[2]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[6],na),&rect[3]) != TCL_OK))
	    croak("bad canvas coordinate for canvas enclosed search");
	return TkpFindArea(canvasPtr, sp, rect, uid, 1);
    } else if ((c == 'o') && (strncmp(searchcom, "overlapping", length) == 0)) {
	double rect[4];
	
	if (items != 6)
	    croak("wrong # args for searchspec overlapping");

	if ((TkGetCanvasCoord(canvasPtr, SvPV(st[3], na), &rect[0]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[4],na),&rect[1]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[5],na),&rect[2]) != TCL_OK)
	    || (TkGetCanvasCoord(canvasPtr,SvPV(st[6],na),&rect[3]) != TCL_OK))
	    croak("bad canvas coordinate for searchspec enclosed");

	return TkpFindArea(canvasPtr, sp, rect, uid, 0);
    } else if ((c == 'w') && (strncmp(searchcom, "withtag", length) == 0)) {
	int pushedany = 0;

	if (items != 3)
	    croak("wrong # args for canvas item withtag search");
	for (itemPtr = StartTagSearch(canvasPtr, SvPV(st[2], na), &search);
		itemPtr != NULL; itemPtr = NextItem(&search)) {
	    if (uid)
		TkpAddTag(itemPtr, uid);
	    else
	    {
		av_store(stack, sp++, sv_2mortal(newSViv(itemPtr->id)));
		pushedany++;
	    }
	}
	if (pushedany)
	{
	    sp--;
	    RECOVERSTACK;
	}
    } else
	croak("bad search command for canvas");
    return sp;
}

/*
 *--------------------------------------------------------------
 *
 * TkpFindArea --
 *
 *	This procedure implements area searches for the "find"
 *	and "addtag" options.
 *
 * Results:
 *	The new top of stack.  If newTag is NULL, then
 *	the ids from all the items overlapping or enclosed
 *	by the rectangle given by rect is pushed on the stack.
 *
 * Side effects:
 *	If uid is non-NULL, then all the items overlapping
 *	or enclosed by the area in rect have that tag added to
 *	their lists of tags.
 *
 *--------------------------------------------------------------
 */

static int
TkpFindArea(canvasPtr, sp, rect, uid, enclosed)
    Tk_Canvas *canvasPtr;		/* Canvas whose items are to be
					 * searched. */
    int sp;
    double *rect;			/* Array of four doubles that
					 * give the coordinates of the
					 * rectangular area to search. */
    Tk_Uid uid;				/* If non-NULL, gives new tag to set
					 * on all found items;  if NULL, then
					 * ids of found items are returned
					 * in interp->result. */
    int enclosed;			/* 0 means overlapping or enclosed
					 * items are OK, 1 means only enclosed
					 * items are OK. */
{
    double tmp;
    int x1, y1, x2, y2;
    register Tk_Item *itemPtr;
    int pushedany = 0;

    if (rect[0] > rect[2]) {
	tmp = rect[0]; rect[0] = rect[2]; rect[2] = tmp;
    }
    if (rect[1] > rect[3]) {
	tmp = rect[1]; rect[1] = rect[3]; rect[3] = tmp;
    }

    /*
     * Use an integer bounding box for a quick test, to avoid
     * calling item-specific code except for items that are close.
     */

    x1 = (rect[0]-1.0);
    y1 = (rect[1]-1.0);
    x2 = (rect[2]+1.0);
    y2 = (rect[3]+1.0);
    for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
	    itemPtr = itemPtr->nextPtr) {
	if ((itemPtr->x1 >= x2) || (itemPtr->x2 <= x1)
		|| (itemPtr->y1 >= y2) || (itemPtr->y2 <= y1)) {
	    continue;
	}
	if ((*itemPtr->typePtr->areaProc)(canvasPtr, itemPtr, rect)
		>= enclosed) {
	    if (uid)
		TkpAddTag(itemPtr, uid);
	    else
	    {
		av_store(stack, sp++, sv_2mortal(newSViv(itemPtr->id)));
		pushedany++;
	    }
	}
    }
    if (pushedany)
    {
	sp--;
	RECOVERSTACK;
    }
    return sp;
}

/*
 *--------------------------------------------------------------
 *
 * RelinkItems --
 *
 *	Move one or more items to a different place in the
 *	display order for a canvas.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The items identified by "tag" are moved so that they
 *	are all together in the display list and immediately
 *	after prevPtr.  The order of the moved items relative
 *	to each other is not changed.
 *
 *--------------------------------------------------------------
 */

static void
RelinkItems(canvasPtr, tag, prevPtr)
    Tk_Canvas *canvasPtr;	/* Canvas to be modified. */
    char *tag;			/* Tag identifying items to be moved
				 * in the redisplay list. */
    Tk_Item *prevPtr;		/* Reposition the items so that they
				 * go just after this item (NULL means
				 * put at beginning of list). */
{
    register Tk_Item *itemPtr;
    TagSearch search;
    Tk_Item *firstMovePtr, *lastMovePtr;

    /*
     * Find all of the items to be moved and remove them from
     * the list, making an auxiliary list running from firstMovePtr
     * to lastMovePtr.  Record their areas for redisplay.
     */

    firstMovePtr = lastMovePtr = NULL;
    for (itemPtr = StartTagSearch(canvasPtr, tag, &search);
	    itemPtr != NULL; itemPtr = NextItem(&search)) {
	if (itemPtr == prevPtr) {
	    /*
	     * Item after which insertion is to occur is being
	     * moved!  Switch to insert after its predecessor.
	     */

	    prevPtr = search.prevPtr;
	}
	if (search.prevPtr == NULL) {
	    canvasPtr->firstItemPtr = itemPtr->nextPtr;
	} else {
	    search.prevPtr->nextPtr = itemPtr->nextPtr;
	}
	if (canvasPtr->lastItemPtr == itemPtr) {
	    canvasPtr->lastItemPtr = search.prevPtr;
	}
	if (firstMovePtr == NULL) {
	    firstMovePtr = itemPtr;
	} else {
	    lastMovePtr->nextPtr = itemPtr;
	}
	lastMovePtr = itemPtr;
	EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		itemPtr->x2, itemPtr->y2);
	canvasPtr->flags |= REPICK_NEEDED;
    }

    /*
     * Insert the list of to-be-moved items back into the canvas's
     * at the desired position.
     */

    if (firstMovePtr == NULL) {
	return;
    }
    if (prevPtr == NULL) {
	lastMovePtr->nextPtr = canvasPtr->firstItemPtr;
	canvasPtr->firstItemPtr = firstMovePtr;
    } else {
	lastMovePtr->nextPtr = prevPtr->nextPtr;
	prevPtr->nextPtr = firstMovePtr;
    }
    if (canvasPtr->lastItemPtr == prevPtr) {
	canvasPtr->lastItemPtr = lastMovePtr;
    }
}

/*
 *--------------------------------------------------------------
 *
 * CanvasBindProc --
 *
 *	This procedure is invoked by the Tk dispatcher to handle
 *	events associated with bindings on items.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Depends on the command invoked as part of the binding
 *	(if there was any).
 *
 *--------------------------------------------------------------
 */

static void
CanvasBindProc(clientData, eventPtr)
    ClientData clientData;		/* Pointer to canvas structure. */
    XEvent *eventPtr;			/* Pointer to X event that just
					 * happened. */
{
    Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;

    Tk_Preserve((ClientData) canvasPtr);

    /*
     * This code below keeps track of the current modifier state in
     * canvasPtr>state.  This information is used to defer repicks of
     * the current item while buttons are down.
     */

    if ((eventPtr->type == ButtonPress) || (eventPtr->type == ButtonRelease)) {
	int mask;

	switch (eventPtr->xbutton.button) {
	    case Button1:
		mask = Button1Mask;
		break;
	    case Button2:
		mask = Button2Mask;
		break;
	    case Button3:
		mask = Button3Mask;
		break;
	    case Button4:
		mask = Button4Mask;
		break;
	    case Button5:
		mask = Button5Mask;
		break;
	    default:
		mask = 0;
		break;
	}
	CanvasDoEvent(canvasPtr, eventPtr);
	eventPtr->xbutton.state ^= mask;
	canvasPtr->state = eventPtr->xbutton.state;
	PickCurrentItem(canvasPtr, eventPtr);
	eventPtr->xbutton.state ^= mask;
	goto done;
    } else if ((eventPtr->type == EnterNotify)
	    || (eventPtr->type == LeaveNotify)) {
	canvasPtr->state = eventPtr->xcrossing.state;
	PickCurrentItem(canvasPtr, eventPtr);
	goto done;
    } else if (eventPtr->type == MotionNotify) {
	canvasPtr->state = eventPtr->xmotion.state;
	PickCurrentItem(canvasPtr, eventPtr);
    }
    CanvasDoEvent(canvasPtr, eventPtr);

    done:
    Tk_Release((ClientData) canvasPtr);
}

/*
 *--------------------------------------------------------------
 *
 * PickCurrentItem --
 *
 *	Find the topmost item in a canvas that contains a given
 *	location and mark the the current item.  If the current
 *	item has changed, generate a fake exit event on the old
 *	current item and a fake enter event on the new current
 *	item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The current item for canvasPtr may change.  If it does,
 *	then the commands associated with item entry and exit
 *	could do just about anything.
 *
 *--------------------------------------------------------------
 */

static void
PickCurrentItem(canvasPtr, eventPtr)
    register Tk_Canvas *canvasPtr;	/* Canvas pointer in which to select
					 * current item. */
    XEvent *eventPtr;			/* Event describing location of
					 * mouse cursor.  Must be EnterWindow,
					 * LeaveWindow, ButtonRelease, or
					 * MotionNotify. */
{
    Tk_Item *closestPtr = NULL;

    /*
     * If a button is down, then don't do anything at all;  we'll be
     * called again when all buttons are up, and we can repick then.
     * This implements a form of mouse grabbing for canvases.
     */

    if (canvasPtr->state
	    & (Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask)) {
	return;
    }

    /*
     * Save information about this event in the canvas.  The event in
     * the canvas is used for two purposes:
     *
     * 1. Event bindings: if the current item changes, fake events are
     *    generated to allow item-enter and item-leave bindings to trigger.
     * 2. Reselection: if the current item gets deleted, can use the
     *    saved event to find a new current item.
     * Translate MotionNotify events into EnterNotify events, since that's
     * what gets reported to item handlers.
     */

    if (eventPtr != &canvasPtr->pickEvent) {
	if ((eventPtr->type == MotionNotify)
		|| (eventPtr->type == ButtonRelease)) {
	    canvasPtr->pickEvent.xcrossing.type = EnterNotify;
	    canvasPtr->pickEvent.xcrossing.serial = eventPtr->xmotion.serial;
	    canvasPtr->pickEvent.xcrossing.send_event
		    = eventPtr->xmotion.send_event;
	    canvasPtr->pickEvent.xcrossing.display = eventPtr->xmotion.display;
	    canvasPtr->pickEvent.xcrossing.window = eventPtr->xmotion.window;
	    canvasPtr->pickEvent.xcrossing.root = eventPtr->xmotion.root;
	    canvasPtr->pickEvent.xcrossing.subwindow = None;
	    canvasPtr->pickEvent.xcrossing.time = eventPtr->xmotion.time;
	    canvasPtr->pickEvent.xcrossing.x = eventPtr->xmotion.x;
	    canvasPtr->pickEvent.xcrossing.y = eventPtr->xmotion.y;
	    canvasPtr->pickEvent.xcrossing.x_root = eventPtr->xmotion.x_root;
	    canvasPtr->pickEvent.xcrossing.y_root = eventPtr->xmotion.y_root;
	    canvasPtr->pickEvent.xcrossing.mode = NotifyNormal;
	    canvasPtr->pickEvent.xcrossing.detail = NotifyNonlinear;
	    canvasPtr->pickEvent.xcrossing.same_screen
		    = eventPtr->xmotion.same_screen;
	    canvasPtr->pickEvent.xcrossing.focus = False;
	    canvasPtr->pickEvent.xcrossing.state = eventPtr->xmotion.state;
	} else  {
	    canvasPtr->pickEvent = *eventPtr;
	}
    }

    /*
     * A LeaveNotify event automatically means that there's no current
     * object, so the rest of the code below can be skipped.
     */

    if (canvasPtr->pickEvent.type != LeaveNotify) {
	int x1, y1, x2, y2;
	double coords[2];
	register Tk_Item *itemPtr;

	coords[0] = canvasPtr->pickEvent.xcrossing.x + canvasPtr->xOrigin;
	coords[1] = canvasPtr->pickEvent.xcrossing.y + canvasPtr->yOrigin;
	x1 = coords[0] - canvasPtr->closeEnough;
	y1 = coords[1] - canvasPtr->closeEnough;
	x2 = coords[0] + canvasPtr->closeEnough;
	y2 = coords[1] + canvasPtr->closeEnough;

	for (itemPtr = canvasPtr->firstItemPtr; itemPtr != NULL;
		itemPtr = itemPtr->nextPtr) {
	    if ((itemPtr->x1 > x2) || (itemPtr->x2 < x1)
		    || (itemPtr->y1 > y2) || (itemPtr->y2 < y1)) {
		continue;
	    }
	    if ((*itemPtr->typePtr->pointProc)(canvasPtr,
		    itemPtr, coords) <= canvasPtr->closeEnough) {
		closestPtr = itemPtr;
	    }
	}
    }

    /*
     * Simulate a LeaveNotify event on the previous current item and
     * an EnterNotify event on the new current item.  Remove the "current"
     * tag from the previous current item and place it on the new current
     * item.
     */

    if (closestPtr == canvasPtr->currentItemPtr) {
	return;
    }
    if (canvasPtr->currentItemPtr != NULL) {
	XEvent event;
	Tk_Item *itemPtr = canvasPtr->currentItemPtr;
	int i;

	event = canvasPtr->pickEvent;
	event.type = LeaveNotify;
	CanvasDoEvent(canvasPtr, &event);

	/*
	 * The check below is needed because there could be an event
	 * handler for <LeaveNotify> that deletes the current item.
	 */

	if (itemPtr == canvasPtr->currentItemPtr) {
	    for (i = itemPtr->numTags-1; i >= 0; i--) {
		if (itemPtr->tagPtr[i] == currentUid) {
		    itemPtr->tagPtr[i] = itemPtr->tagPtr[itemPtr->numTags-1];
		    itemPtr->numTags--;
		    break;
		}
	    }
	}
    }
    canvasPtr->currentItemPtr = closestPtr;
    if (canvasPtr->currentItemPtr != NULL) {
	XEvent event;

	TkpAddTag(closestPtr, currentUid);
	event = canvasPtr->pickEvent;
	event.type = EnterNotify;
	CanvasDoEvent(canvasPtr, &event);
    }
}

/*
 *--------------------------------------------------------------
 *
 * CanvasDoEvent --
 *
 *	This procedure is called to invoke binding processing
 *	for a new event that is associated with the current item
 *	for a canvas.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Depends on the bindings for the canvas.
 *
 *--------------------------------------------------------------
 */

static void
CanvasDoEvent(canvasPtr, eventPtr)
    Tk_Canvas *canvasPtr;		/* Canvas widget in which event
					 * occurred. */
    XEvent *eventPtr;			/* Real or simulated X event that
					 * is to be processed. */
{
#define NUM_STATIC 3
    ClientData staticObjects[NUM_STATIC];
    ClientData *objectPtr;
    int numObjects, i;
    register Tk_Item *itemPtr;

    if (canvasPtr->bindingTable == NULL) {
	return;
    }

    itemPtr = canvasPtr->currentItemPtr;
    if ((eventPtr->type == KeyPress) || (eventPtr->type == KeyRelease)) {
	itemPtr = canvasPtr->focusItemPtr;
    }
    if (itemPtr == NULL) {
	return;
    }

    /*
     * Set up an array with all the relevant objects for processing
     * this event.  The relevant objects are (a) the event's item,
     * (b) the tags associated with the event's item, and (c) the
     * tag "all".  If there are a lot of tags then malloc an array
     * to hold all of the objects.
     */

    numObjects = itemPtr->numTags + 2;
    if (numObjects <= NUM_STATIC) {
	objectPtr = staticObjects;
    } else {
	objectPtr = (ClientData *) ckalloc((unsigned)
		(numObjects * sizeof(ClientData)));
    }
    objectPtr[0] = (ClientData) itemPtr;
    for (i = itemPtr->numTags-1; i >= 0; i--) {
	objectPtr[i+1] = (ClientData) itemPtr->tagPtr[i];
    }
    objectPtr[itemPtr->numTags+1] = (ClientData) allUid;

    /*
     * Invoke the binding system, then free up the object array if
     * it was malloc-ed.
     */

    TkpBindEvent(canvasPtr->bindingTable, eventPtr, canvasPtr->tkwin,
	    numObjects, objectPtr);
    if (objectPtr != staticObjects) {
	ckfree((char *) objectPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CanvasBlinkProc --
 *
 *	This procedure is called as a timer handler to blink the
 *	insertion cursor off and on.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The cursor gets turned on or off, redisplay gets invoked,
 *	and this procedure reschedules itself.
 *
 *----------------------------------------------------------------------
 */

static void
CanvasBlinkProc(clientData)
    ClientData clientData;	/* Pointer to record describing entry. */
{
    register Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;

    if (!(canvasPtr->flags & GOT_FOCUS) || (canvasPtr->insertOffTime == 0)) {
	return;
    }
    if (canvasPtr->flags & CURSOR_ON) {
	canvasPtr->flags &= ~CURSOR_ON;
	canvasPtr->insertBlinkHandler = Tk_CreateTimerHandler(
		canvasPtr->insertOffTime, CanvasBlinkProc,
		(ClientData) canvasPtr);
    } else {
	canvasPtr->flags |= CURSOR_ON;
	canvasPtr->insertBlinkHandler = Tk_CreateTimerHandler(
		canvasPtr->insertOffTime, CanvasBlinkProc,
		(ClientData) canvasPtr);
    }
    if (canvasPtr->focusItemPtr != NULL) {
	EventuallyRedrawArea(canvasPtr, canvasPtr->focusItemPtr->x1,
		canvasPtr->focusItemPtr->y1, canvasPtr->focusItemPtr->x2,
		canvasPtr->focusItemPtr->y2);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CanvasFocusProc --
 *
 *	This procedure is called whenever a canvas gets or loses the
 *	input focus.  It's also called whenever the window is
 *	reconfigured while it has the focus.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The cursor gets turned on or off.
 *
 *----------------------------------------------------------------------
 */

static void
CanvasFocusProc(canvasPtr, gotFocus)
    Tk_Canvas *canvasPtr;	/* Canvas that just got or lost focus. */
    int gotFocus;		/* 1 means window is getting focus, 0 means
				 * it's losing it. */
{
    Tk_DeleteTimerHandler(canvasPtr->insertBlinkHandler);
    if (gotFocus) {
	canvasPtr->flags |= GOT_FOCUS | CURSOR_ON;
	if (canvasPtr->insertOffTime != 0) {
	    canvasPtr->insertBlinkHandler = Tk_CreateTimerHandler(
		    canvasPtr->insertOffTime, CanvasBlinkProc,
		    (ClientData) canvasPtr);
	}
    } else {
	canvasPtr->flags &= ~(GOT_FOCUS | CURSOR_ON);
	canvasPtr->insertBlinkHandler = (Tk_TimerToken) NULL;
    }
    if (canvasPtr->focusItemPtr != NULL) {
	EventuallyRedrawArea(canvasPtr, canvasPtr->focusItemPtr->x1,
		canvasPtr->focusItemPtr->y1, canvasPtr->focusItemPtr->x2,
		canvasPtr->focusItemPtr->y2);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CanvasSelectTo --
 *
 *	Modify the selection by moving its un-anchored end.  This could
 *	make the selection either larger or smaller.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The selection changes.
 *
 *----------------------------------------------------------------------
 */

static void
CanvasSelectTo(canvasPtr, itemPtr, index)
    register Tk_Canvas *canvasPtr;		/* Information about widget. */
    register Tk_Item *itemPtr;		/* Item that is to hold selection. */
    int index;				/* Index of element that is to
					 * become the "other" end of the
					 * selection. */
{
    int oldFirst, oldLast;
    Tk_Item *oldSelPtr;

    oldFirst = canvasPtr->selectFirst;
    oldLast = canvasPtr->selectLast;
    oldSelPtr = canvasPtr->selItemPtr;

    /*
     * Grab the selection if we don't own it already.
     */

    if (canvasPtr->selItemPtr == NULL) {
	Tk_OwnSelection(canvasPtr->tkwin, CanvasLostSelection,
		(ClientData) canvasPtr);
    } else if (canvasPtr->selItemPtr != itemPtr) {
	EventuallyRedrawArea(canvasPtr, canvasPtr->selItemPtr->x1,
		canvasPtr->selItemPtr->y1, canvasPtr->selItemPtr->x2,
		canvasPtr->selItemPtr->y2);
    }
    canvasPtr->selItemPtr = itemPtr;

    if (canvasPtr->anchorItemPtr != itemPtr) {
	canvasPtr->anchorItemPtr = itemPtr;
	canvasPtr->selectAnchor = index;
    }
    if (canvasPtr->selectAnchor <= index) {
	canvasPtr->selectFirst = canvasPtr->selectAnchor;
	canvasPtr->selectLast = index;
    } else {
	canvasPtr->selectFirst = index;
	canvasPtr->selectLast = canvasPtr->selectAnchor - 1;
    }
    if ((canvasPtr->selectFirst != oldFirst)
	    || (canvasPtr->selectLast != oldLast)
	    || (itemPtr != oldSelPtr)) {
	EventuallyRedrawArea(canvasPtr, itemPtr->x1, itemPtr->y1,
		itemPtr->x2, itemPtr->y2);
    }
}

/*
 *--------------------------------------------------------------
 *
 * CanvasFetchSelection --
 *
 *	This procedure is invoked by Tk to return part or all of
 *	the selection, when the selection is in a canvas widget.
 *	This procedure always returns the selection as a STRING.
 *
 * Results:
 *	The return value is the number of non-NULL bytes stored
 *	at buffer.  Buffer is filled (or partially filled) with a
 *	NULL-terminated string containing part or all of the selection,
 *	as given by offset and maxBytes.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
CanvasFetchSelection(clientData, offset, buffer, maxBytes)
    ClientData clientData;		/* Information about canvas widget. */
    int offset;				/* Offset within selection of first
					 * character to be returned. */
    char *buffer;			/* Location in which to place
					 * selection. */
    int maxBytes;			/* Maximum number of bytes to place
					 * at buffer, not including terminating
					 * NULL character. */
{
    register Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;

    if (canvasPtr->selItemPtr == NULL) {
	return -1;
    }
    if (canvasPtr->selItemPtr->typePtr->selectionProc == NULL) {
	return -1;
    }
    return (*canvasPtr->selItemPtr->typePtr->selectionProc)(
	    canvasPtr, canvasPtr->selItemPtr, offset, buffer, maxBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * CanvasLostSelection --
 *
 *	This procedure is called back by Tk when the selection is
 *	grabbed away from a canvas widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The existing selection is unhighlighted, and the window is
 *	marked as not containing a selection.
 *
 *----------------------------------------------------------------------
 */

static void
CanvasLostSelection(clientData)
    ClientData clientData;		/* Information about entry widget. */
{
    Tk_Canvas *canvasPtr = (Tk_Canvas *) clientData;

    if (canvasPtr->selItemPtr != NULL) {
	EventuallyRedrawArea(canvasPtr, canvasPtr->selItemPtr->x1,
		canvasPtr->selItemPtr->y1, canvasPtr->selItemPtr->x2,
		canvasPtr->selItemPtr->y2);
    }
    canvasPtr->selItemPtr = NULL;
}

/*
 *--------------------------------------------------------------
 *
 * TkGetCanvasCoord --
 *
 *	Given a string, returns a floating-point canvas coordinate
 *	corresponding to that string.
 *
 * Results:
 *	The return value is a standard Tcl return result.  If
 *	TCL_OK is returned, then everything went well and the
 *	canvas coordinate is stored at *doublePtr;  otherwise
 *	TCL_ERROR is returned and an error message is left in
 *	canvasPtr->interp->result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
TkGetCanvasCoord(canvasPtr, string, doublePtr)
    Tk_Canvas *canvasPtr;	/* Canvas to which coordinate applies. */
    char *string;		/* Describes coordinate (any screen
				 * coordinate form may be used here). */
    double *doublePtr;		/* Place to store converted coordinate. */
{
    if (Tk_GetScreenMM(tclinterp, canvasPtr->tkwin, string,
	    doublePtr) != TCL_OK) {
	return TCL_ERROR;
    }
    *doublePtr *= canvasPtr->pixelsPerMM;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * GridAlign --
 *
 *	Given a coordinate and a grid spacing, this procedure
 *	computes the location of the nearest grid line to the
 *	coordinate.
 *
 * Results:
 *	The return value is the location of the grid line nearest
 *	to coord.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static double
GridAlign(coord, spacing)
    double coord;		/* Coordinate to grid-align. */
    double spacing;		/* Spacing between grid lines.   If <= 0
				 * then no alignment is done. */
{
    if (spacing <= 0.0) {
	return coord;
    }
    if (coord < 0) {
	return -((int) ((-coord)/spacing + 0.5)) * spacing;
    }
    return ((int) (coord/spacing + 0.5)) * spacing;
}

/*
 *--------------------------------------------------------------
 *
 * CanvasUpdateScrollbars --
 *
 *	This procedure is invoked whenever a canvas has changed in
 *	a way that requires scrollbars to be redisplayed (e.g. the
 *	view in the canvas has changed).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If there are scrollbars associated with the canvas, then
 *	their scrolling commands are invoked to cause them to
 *	redisplay.  If errors occur, additional Tcl commands may
 *	be invoked to process the errors.
 *
 *--------------------------------------------------------------
 */

static void
CanvasUpdateScrollbars(canvasPtr)
    register Tk_Canvas *canvasPtr;		/* Information about canvas. */
{
    int size, first, last, page;
    int sp = stack_sp - stack_base;

#define ROUND(number)						\
    if (number >= 0) {						\
	number = (number + canvasPtr->scrollIncrement/2)	\
		/canvasPtr->scrollIncrement;			\
    } else {							\
	number = -(((-number) + canvasPtr->scrollIncrement/2)	\
		/canvasPtr->scrollIncrement);			\
    }

    canvasPtr->flags &= ~UPDATE_SCROLLBARS;
    
    if (CALLBACKOK(&canvasPtr->xScrollCallback)) {
	size = ((canvasPtr->scrollX2 - canvasPtr->scrollX1)
		/canvasPtr->scrollIncrement) + 1;
	first = canvasPtr->xOrigin - canvasPtr->scrollX1;
	ROUND(first);
	last = canvasPtr->xOrigin + Tk_Width(canvasPtr->tkwin)
		- 1 - canvasPtr->scrollX1;
	ROUND(last);
	page = last - first - 1;
	if (page <= 0) {
	    page = 1;
	}
	av_store(stack, sp + 2, sv_2mortal(newSViv(size)));
	av_store(stack, sp + 3, sv_2mortal(newSViv(page)));
	av_store(stack, sp + 4, sv_2mortal(newSViv(first)));
	av_store(stack, sp + 5, sv_2mortal(newSViv(last)));
	RECOVERSTACK;
	TkpCallback(&canvasPtr->xScrollCallback, 4);
    }

    if (CALLBACKOK(&canvasPtr->yScrollCallback)) {
	size = ((canvasPtr->scrollY2 - canvasPtr->scrollY1)
		/canvasPtr->scrollIncrement) + 1;
	first = canvasPtr->yOrigin - canvasPtr->scrollY1;
	ROUND(first);
	last = canvasPtr->yOrigin + Tk_Height(canvasPtr->tkwin)
		- 1 - canvasPtr->scrollY1;
	ROUND(last);
	page = last - first - 1;
	if (page <= 0) {
	    page = 1;
	}
	av_store(stack, sp + 2, sv_2mortal(newSViv(size)));
	av_store(stack, sp + 3, sv_2mortal(newSViv(page)));
	av_store(stack, sp + 4, sv_2mortal(newSViv(first)));
	av_store(stack, sp + 5, sv_2mortal(newSViv(last)));
	RECOVERSTACK;
	TkpCallback(&canvasPtr->yScrollCallback, 4);
    }
}

/*
 *--------------------------------------------------------------
 *
 * CanvasSetOrigin --
 *
 *	This procedure is invoked to change the mapping between
 *	canvas coordinates and screen coordinates in the canvas
 *	window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The canvas will be redisplayed to reflect the change in
 *	view.  In addition, scrollbars will be updated if there
 *	are any.
 *
 *--------------------------------------------------------------
 */

static void
CanvasSetOrigin(canvasPtr, xOrigin, yOrigin)
    register Tk_Canvas *canvasPtr;	/* Information about canvas. */
    int xOrigin;			/* New X origin for canvas (canvas
					 * x-coord corresponding to left edge
					 * of canvas window). */
    int yOrigin;			/* New Y origin for canvas (canvas
					 * y-coord corresponding to top edge
					 * of canvas window). */
{
    int left, right, top, bottom;

    /*
     * Adjust the origin if necessary to keep as much as possible of the
     * canvas in the view.  The variables left, right, etc. keep track of
     * how much extra space there is on each side of the view before it
     * will stick out past the scroll region.  If one side sticks out past
     * the edge of the scroll region, adjust the view to bring that side
     * back to the edge of the scrollregion (but don't move it so much that
     * the other side sticks out now).
     */

    if ((canvasPtr->confine) && (canvasPtr->regionString != NULL)) {
	left = xOrigin - canvasPtr->scrollX1;
	right = canvasPtr->scrollX2 - (xOrigin + Tk_Width(canvasPtr->tkwin));
	top = yOrigin - canvasPtr->scrollY1;
	bottom = canvasPtr->scrollY2 - (yOrigin + Tk_Height(canvasPtr->tkwin));
	if ((left < 0) && (right > 0)) {
	    xOrigin += (right > -left) ? -left : right;
	} else if ((right < 0) && (left > 0)) {
	    xOrigin -= (left > -right) ? -right : left;
	}
	if ((top < 0) && (bottom > 0)) {
	    yOrigin += (bottom > -top) ? -top : bottom;
	} else if ((bottom < 0) && (top > 0)) {
	    yOrigin -= (top > -bottom) ? -bottom : top;
	}
    }

    if ((xOrigin == canvasPtr->xOrigin) && (yOrigin == canvasPtr->yOrigin)) {
	return;
    }

    /*
     * Tricky point:  must redisplay not only everything that's visible
     * in the window's final configuration, but also everything that was
     * visible in the initial configuration.  This is needed because some
     * item types, like windows, need to know when they move off-screen
     * so they can explicitly undisplay themselves.
     */

    EventuallyRedrawArea(canvasPtr, canvasPtr->xOrigin, canvasPtr->yOrigin,
	    canvasPtr->xOrigin + Tk_Width(canvasPtr->tkwin),
	    canvasPtr->yOrigin + Tk_Height(canvasPtr->tkwin));
    canvasPtr->xOrigin = xOrigin;
    canvasPtr->yOrigin = yOrigin;
    canvasPtr->flags |= UPDATE_SCROLLBARS;
    EventuallyRedrawArea(canvasPtr, canvasPtr->xOrigin, canvasPtr->yOrigin,
	    canvasPtr->xOrigin + Tk_Width(canvasPtr->tkwin),
	    canvasPtr->yOrigin + Tk_Height(canvasPtr->tkwin));
}

/*
 *--------------------------------------------------------------
 *
 * CanvasTagsParseProc --
 *
 *	This procedure is invoked during option processing to handle
 *	"-tags" options for canvas items.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	The tags for a given item get replaced by those indicated
 *	in the value argument.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static int
CanvasTagsParseProc(clientData, interp, tkwin, value, widgRec, offset)
    ClientData clientData;		/* Not used.*/
    Tcl_Interp *interp;			/* Used for reporting errors. */
    Tk_Window tkwin;			/* Window containing canvas widget. */
    char *value;			/* Value of option (list of tag
					 * names). */
    char *widgRec;			/* Pointer to record for item. */
    int offset;				/* Offset into item (ignored). */
{
    register Tk_Item *itemPtr = (Tk_Item *) widgRec;
    int argc, i;
    char **argv;
    Tk_Uid *newPtr;

    /*
     * Break the value up into the individual tag names.
     */

    if (Tcl_SplitList(interp, value, &argc, &argv) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Make sure that there's enough space in the item to hold the
     * tag names.
     */

    if (itemPtr->tagSpace < argc) {
	newPtr = (Tk_Uid *) ckalloc((unsigned) (argc * sizeof(Tk_Uid)));
	for (i = itemPtr->numTags-1; i >= 0; i--) {
	    newPtr[i] = itemPtr->tagPtr[i];
	}
	if (itemPtr->tagPtr != itemPtr->staticTagSpace) {
	    ckfree((char *) itemPtr->tagPtr);
	}
	itemPtr->tagPtr = newPtr;
	itemPtr->tagSpace = argc;
    }
    itemPtr->numTags = argc;
    for (i = 0; i < argc; i++) {
	itemPtr->tagPtr[i] = Tk_GetUid(argv[i]);
    }
    ckfree((char *) argv);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * CanvasTagsPrintProc --
 *
 *	This procedure is invoked by the Tk configuration code
 *	to produce a printable string for the "-tags" configuration
 *	option for canvas items.
 *
 * Results:
 *	The return value is a string describing all the tags for
 *	the item referred to by "widgRec".  In addition, *freeProcPtr
 *	is filled in with the address of a procedure to call to free
 *	the result string when it's no longer needed (or NULL to
 *	indicate that the string doesn't need to be freed).
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static char *
CanvasTagsPrintProc(clientData, tkwin, widgRec, offset, freeProcPtr)
    ClientData clientData;		/* Ignored. */
    Tk_Window tkwin;			/* Window containing canvas widget. */
    char *widgRec;			/* Pointer to record for item. */
    int offset;				/* Ignored. */
    Tcl_FreeProc **freeProcPtr;		/* Pointer to variable to fill in with
					 * information about how to reclaim
					 * storage for return string. */
{
    register Tk_Item *itemPtr = (Tk_Item *) widgRec;

    if (itemPtr->numTags == 0) {
	*freeProcPtr = (Tcl_FreeProc *) NULL;
	return "";
    }
    if (itemPtr->numTags == 1) {
	*freeProcPtr = (Tcl_FreeProc *) NULL;
	return (char *) itemPtr->tagPtr[0];
    }
    *freeProcPtr = (Tcl_FreeProc *) free;
    return Tcl_Merge(itemPtr->numTags, (char **) itemPtr->tagPtr);
}

void bootC_Canvas _((void))
{
    char *file = __FILE__;

    canvasClassRec.package = "Canvas";
    canvasClassRec.stash = Nullhv; /* filled in by init_widget */
    canvasClassRec.newwidgetproc = TkpCanvasCmd;
    canvasClassRec.newix = 0;
    canvasClassRec.widgetcmdproc = TkpCanvasWidgetCmd;
    canvasClassRec.methods = canvasMethods;
    canvasClassRec.filename = file;
    init_widget(&canvasClassRec);
}
